import os
import click
import importlib.metadata
from getpass import getuser
from click import echo, secho
from .utils import extract_project_template, make_python_identifier


@click.group()
def jembe():
    pass


@jembe.command()
@click.option(
    "--name",
    help="Name",
    prompt="Project Name",
    required=True,
    default=lambda: make_python_identifier(os.path.basename(os.getcwd())),
)
@click.option(
    "--description", help="Description", prompt=True, required=False, default=""
)
def startproject( name, description):
    """Starts new Jembe project in current directory"""
    name = make_python_identifier(name)
    ctx = dict(
        project_name=name,
        project_description=description,
        project_author=getuser(),
        jembe_version=importlib.metadata.version("jembe"),
        secret_key=str(os.urandom(24).__repr__()),
    )
    extract_project_template("basic", ctx=ctx)

    echo()
    echo("New project is suceessfully created in current directory!", color=True)
    echo()
    echo("To install required development dependencies execute:")
    secho("\t$ pip install -e .[dev]", bold=True)
    echo()
    echo("To start development web server execute:")
    secho("\t$ flask run", bold=True)
    echo()
    echo("To package project for deployment run:")
    secho("\t$ python -m build", bold=True)
    echo()



from typing import TYPE_CHECKING, Optional
from jembe import Component
from flask import json
from jembe import action

# from jembe import action, listener
# from jembe.utils import build_url

if TYPE_CHECKING:
    from uuid import UUID

    # from jembe import Event


def test_simple_page(jmb, client):
    class SimplePage(Component):
        pass

    jmb.add_page("simple_page", SimplePage)
    r = client.get("/simple_page")
    assert r.status_code == 200
    assert b"<h1>Simple page</h1>" in r.data
    assert (
        b'<html lang="en" jmb-name="/simple_page" jmb-data=\'{"actions":{},"changesUrl":true,"state":{},"url":"/simple_page"}\''
        in r.data
    )


def test_empty_page(jmb, client):
    class EmptyPage(Component):
        def display(self):
            return self.render_template_string("")

    jmb.add_page("page", EmptyPage)
    r = client.get("/page")
    assert r.status_code == 200
    assert (
        b'<html jmb-name="/page" jmb-data=\'{"actions":{},"changesUrl":true,"state":{},"url":"/page"}\'>'
        in r.data
    )
    assert b"<div></div>" in r.data


def test_simple_page_with_custom_template(jmb, client):
    class SimplePage(Component):
        pass

    jmb.add_page(
        "simple_page",
        SimplePage,
        SimplePage.Config(template="simple_page_custom.html"),
    )

    r = client.get("/simple_page")
    assert r.status_code == 200
    assert b"<h1>Custom simple page</h1>" in r.data


def test_page_with_template(jmb, client):
    class Page(Component):
        class Config(Component.Config):
            def __init__(self, template="simple_page_custom.html", **params):
                super().__init__(template=template, **params)

    jmb.add_page("page", Page)
    r = client.get("/page")
    assert r.status_code == 200
    assert b"<h1>Custom simple page</h1>" in r.data


def test_page_with_decorator(jmb, client):
    @jmb.page("welcome")
    class WelcomePage(Component):
        pass

    @jmb.page("hello", Component.Config(template="simple_page.html"))
    class HelloPage(Component):
        pass

    r = client.get("/welcome")
    assert r.status_code == 200
    assert b"<h1>Welcome</h1>" in r.data

    r = client.get("/hello")
    assert r.status_code == 200
    assert b"<h1>Simple page</h1>" in r.data


def test_blog_post_page(jmb, client):
    @jmb.page("blogpost")
    class BlogPost(Component):
        def __init__(self, id: int):
            self.id = id
            super().__init__()

        def display(self):
            return self.render_template_string(
                """
                <html>
                <head></head>
                <body>Blog post id == {{id}}</body>
                </html>
                """
            )

    r = client.get("/blogpost/1")
    assert r.status_code == 200
    assert b"Blog post id == 1" in r.data


def test_counter_on_page(jmb, client):
    @jmb.page("cop")
    class CounterOnPage(Component):
        def __init__(self, counter: int = 0):
            super().__init__()

        @action
        def increase(self):
            self.state.counter += 1

        def display(self):
            return self.render_template_string(
                """<html>
                <head></head>
                <body>
                    <div>Counter: {{counter}}</div>
                    <div>c={{state.counter}}</div>
                    <button jmb:click="increase()"/>
                </body>
                </html>"""
            )

    # First load
    res = client.get("/cop")
    assert res.status_code == 200
    assert b"Counter: 0" in res.data
    assert b"c=0" in res.data
    assert (
        b'jmb-name="/cop" jmb-data=\'{"actions":{"increase":true},"changesUrl":true,"state":{"counter":0},"url":"/cop"}\''
        in res.data
    )
    assert b'<button jmb:click="increase()"' in res.data

    # Refresh page with ajax request
    ajax_post_data = json.dumps(
        dict(
            components=[],
            commands=[
                dict(
                    type="call",  # emit, initialise
                    componentExecName="/cop",
                    actionName="display",
                    args=list(),
                    kwargs=dict(),
                )
            ],
        )
    )
    r = client.post("/cop", data=ajax_post_data, headers={"x-jembe": True})
    assert r.status_code == 200

    ajax_response_data = json.loads(r.data)
    assert len(ajax_response_data) == 1
    assert ajax_response_data[0]["dom"].startswith("<html")
    assert ajax_response_data[0]["dom"].endswith("</html>")
    assert "Counter: 0" in ajax_response_data[0]["dom"]
    assert "c=0" in ajax_response_data[0]["dom"]
    assert '<button jmb:click="increase()"' in ajax_response_data[0]["dom"]
    assert ajax_response_data[0]["execName"] == "/cop"
    assert ajax_response_data[0]["state"] == {"counter": 0}
    assert ajax_response_data[0]["url"] == "/cop"

    # Call increase action
    ajax_post_data = json.dumps(
        dict(
            components=[dict(execName="/cop", state=dict(counter=0))],
            commands=[
                dict(
                    type="call",
                    componentExecName="/cop",
                    actionName="increase",
                    args=list(),
                    kwargs=dict(),
                )
            ],
        )
    )
    r = client.post("/cop", data=ajax_post_data, headers={"x-jembe": True})
    assert r.status_code == 200

    ajax_response_data = json.loads(r.data)
    assert len(ajax_response_data) == 1
    assert "Counter: 1" in ajax_response_data[0]["dom"]
    assert "c=1" in ajax_response_data[0]["dom"]
    assert ajax_response_data[0]["state"] == {"counter": 1}
    assert ajax_response_data[0]["url"] == "/cop"
    assert ajax_response_data[0]["execName"] == "/cop"




from re import T
from typing import (
    Callable,
    Iterable,
    TYPE_CHECKING,
    Optional,
    Type,
    Dict,
    List,
    Tuple,
    Union,
    Sequence,
    NewType,
)
from abc import ABCMeta
from enum import Enum
from itertools import accumulate
from operator import add
from inspect import getmembers, isfunction, signature, Parameter

from jembe.common import import_by_name
from .exceptions import AccessDenied, JembeError
from flask import url_for

if TYPE_CHECKING:  # pragma: no cover
    import inspect
    import jembe
    from .component import ComponentState

UrlPath = NewType("UrlPath", str)


class UrlConvertor(Enum):
    STR0 = "string(minlength=0)"
    STR = "string"
    PATH = "path"
    INT = "int"
    FLOAT = "float"
    UUID = "uuid"


def calc_url_param_identifier(name: str, level: int):
    return name if level == 0 else f"{name}__{level}"


class UrlParamDef:
    def __init__(
        self,
        name: str,
        convertor: Optional["UrlConvertor"] = None,
        identifier: Optional[str] = None,
    ):
        self.name = name
        self.identifier = name if not identifier else identifier
        self.convertor = convertor if convertor else UrlConvertor.STR

    @classmethod
    def from_inspect_parameter(
        cls, param: "inspect.Parameter", level: int
    ) -> "UrlParamDef":
        convertor = UrlConvertor.STR
        if param.annotation.__name__ == "int":
            convertor = UrlConvertor.INT
        elif param.annotation.__name__ == "str":
            convertor = UrlConvertor.STR
        elif param.annotation.__name__ == "float":
            convertor = UrlConvertor.FLOAT
        elif param.annotation.__name__ == "UUID":
            convertor = UrlConvertor.UUID
        elif param.annotation.__name__ == "UrlPath":
            convertor = UrlConvertor.PATH
        return UrlParamDef(
            param.name, convertor, calc_url_param_identifier(param.name, level)
        )

    @property
    def url_pattern(self) -> str:
        return f"<{self.convertor.value}:{self.identifier}>"

    def __repr__(self):
        return self.url_pattern


def action(
    _method=None,
    *,
    deferred=False,
    # deferred_after: Optional[str] = None,
    # deferred_before: Optional[str] = None,
):

    # deferred aciton is executed last after allother actions from parent action template
    # are executed no matter of its postion inside parent action template.

    # Usefull if we need to create breadcrumb or other summary report
    # based from already executed actions

    # deferred_after and deferred_before are used to execute this action after or before
    # other specific deferred action, when multiple actions has deferred execution
    """Decorator to mark Component method as public action

    Public actions can be called from browser using javascript, as follows:

    .. code-block:: html

        <button onclick="jembeClient.component(this).call('action',{param1:value1, param2:value2})">
            Action
        </button>

        Or using Jembe Js ComponentApi:

        <button jmb-on:click="action(value1)">
            Action
        </button>

        or

        <a href="" jmb-on:click.prevent="action()">
            Action
        </a>

    If action is defered than it is executed last, after all other actions from parent action template
    are executed no matter of its postion inside parent action template. This is usefull if we need
    to create breadcrumb or other summary report based from already executed actions.

    Args:
        deferred (bool, optional): If action is deffered. Defaults to False.

    Returns:
        Optional[bool]: Decorated action should return None, True or False
            If decorated action returns:

            - True: then component will redisplay itself.
            - False: then component will NOT redisplay itself.
            - None: then component will redisplay it self based on values of Config.redisplay and state parameters. Defult if will redisplay itself only if state params are changed.
    """
    # This decorator don't anytthing except allow
    # componentconfig.set_compoenent_class to
    # recognise method as action
    def decorator_action(method):
        setattr(method, "_jembe_action", True)
        setattr(method, "_jembe_action_deferred", deferred)
        # setattr(method, "_jembe_action_deferred_after", deferred_after)
        # setattr(method, "_jembe_action_deferred_before", deferred_before)
        return method

    if _method is None:
        return decorator_action
    else:
        return decorator_action(_method)


def listener(
    _method=None,
    *,
    event: Optional[Union[str, Sequence[str]]] = None,
    source: Optional[Union[str, Sequence[str]]] = None,
):
    """Decorator to mark Component method as event listener.

    Event listener listens for emited events.

    Event listener must accepet event parameter like:

    .. code-block:: python

        @listener("_display", source="*")
        def on_display_any_direct_childred(self, event:"jembe.Event"):
            pass

    Args:
        event (Optional[Union[str, Sequence[str]]], optional): Event name to listen for. Defaults to None.

            If event name is:

            - None: it will listen for any event
            - str: it will listen for only event with the same name
            - Sequence[str]: it will listen for any event name from the list of names

        source (Optional[Union[str, Sequence[str]]], optional): Listens for events emited by component specified as source. Defaults to None.

            Listen for:

            - None -> events from any component
            - *                               -> direct children with or without key
            - \*.\*                             -> direct children with any key
            - **/*                            -> all children with or without key
            - **/\*.\*                          -> all children with any key
            - component                       -> match to direct child named "component" without key
            - component.*                     -> match to direct child named "component" with any key
            - component.key                   -> match to direct child named "component with key equals "key"
            - **/component[.[*|<key>]]        -> match to child at any level with "component" name
            - ..                                -> match to parent component
            - ../component[.[*|<key>]]          -> match to sibling
            - /**/.                             -> match to parent at any level
            - /\*\*/component[.[*|<key>]]/\*\*/.    -> match to parent at any level named "component"
            - //                                -> process event from root/page component
            - .                                 -> self

    Returns:
        Optional[bool]: Decorated method should return None, True or False

            If decorated listener returns:

            - True: then component will redisplay itself.
            - False: then component will NOT redisplay itself.
            - None: then component will redisplay it self based on values of Config.redisplay and state parameters. Defult if will redisplay itself only if state params are changed.

    """
    # This decorator don't anytthing except allow
    # componentconfig.set_compoenent_class to
    # recognise method as listener
    def decorator_listener(method):
        setattr(method, "_jembe_listener", True)
        setattr(method, "_jembe_listener_event_name", event)
        setattr(method, "_jembe_listener_source", source)
        return method

    if _method is None:
        return decorator_listener
    else:
        return decorator_listener(_method)


def redisplay(
    _method=None,
    *,
    when_state_changed: Optional[bool] = None,
    when_executed: Optional[bool] = None,
    when_on_page: Optional[bool] = None,
):
    """Decorator for Component.display method to configure when component will be redisplayed

    It's used instead of Component.Config redisplay parameter, when it's more convinient.


    Args:
        when_state_changed (Optional[bool], optional): Redisplay component whent its state is changed. Defaults to None.
        when_executed (Optional[bool], optional): Redisplay component when ever "display" action is called regradles of the component state. Defaults to None.
        when_on_page (Optional[bool], optional): Redisplay component whenever it's on the page. Defaults to None.
    """

    def decorator_action(method):
        flags = set(ComponentConfig.REDISPLAY_DEFAULT_FLAGS)

        def update_flags(state, flag):
            if state == True:
                flags.add(flag)
            elif state == False:
                flags.remove(flag)

        update_flags(when_executed, RedisplayFlag.WHEN_DISPLAY_EXECUTED)
        update_flags(when_state_changed, RedisplayFlag.WHEN_STATE_CHANGED)
        update_flags(when_on_page, RedisplayFlag.WHEN_ON_PAGE)
        setattr(method, "_jembe_redisplay", tuple(flags))
        return method

    if _method is None:
        return decorator_action
    else:
        return decorator_action(_method)


def config(component_config: "jembe.ComponentConfig"):
    """Component decorator to declarativly change Component Configuration

    This decorator changes Component Configuration by suplaying predefined
    params to Component.ComponentConfig.__init__ method.

    .. note::
        ComponentConfig instance in argument is used only to get values of __init__ params.

        Collected params are used when component is registred to Jembe instance,
        allowing parent component to overide configuration of its subcomponents when needed.

    Args:
        component_config (jembe.ComponentConfig): Instance of Component.Config class
    """

    def decorator_class(component_class):
        setattr(
            component_class,
            "_jembe_config_init_params",
            component_config._raw_init_params,
        )
        return component_class

    return decorator_class


class ComponentAction:
    def __init__(
        self,
        name: str,
        deferred: bool = False,
        # deferred_after: Optional[str] = None,
        # deferred_before: Optional[str] = None,
    ):
        self.name = name
        self.deferred = deferred
        # self.deferred_after = deferred_after
        # self.deferred_before = deferred_before

    @classmethod
    def from_method(cls, method) -> "ComponentAction":
        return cls(
            method.__name__,
            getattr(method, "_jembe_action_deferred", False),
            # getattr(method, "_jembe_action_deferred_after", None),
            # getattr(method, "_jembe_action_deferred_before", None),
        )


class ComponentListener:
    def __init__(
        self,
        method_name: str,
        event_name: Union[str, Sequence[str]],
        source: Union[str, Sequence[str]] = (),
    ):
        self.method_name: str = method_name

        self.event_names: Tuple[str, ...] = ()
        if event_name is None:
            pass
        elif isinstance(event_name, str):
            self.event_names = (event_name,)
        else:
            self.event_names = tuple(event_name)

        self.sources: Tuple[str, ...] = ()
        if source is None:
            pass
        elif isinstance(source, str):
            self.sources = (source,)
        else:
            self.sources = tuple(source)

    @classmethod
    def from_method(cls, method) -> "ComponentListener":
        return cls(
            method.__name__,
            getattr(method, "_jembe_listener_event_name", ()),
            getattr(method, "_jembe_listener_source", ()),
        )


class RedisplayFlag(Enum):
    WHEN_STATE_CHANGED = "wsc"
    WHEN_DISPLAY_EXECUTED = "wde"
    WHEN_ON_PAGE = "wop"


def componentConfigInitDecorator(init_method):
    def decoratedInit(self, *args, **kwargs):
        """Saves named init params as self._raw_init_params and apply params from @config"""
        # save original init_params as _raw_init_prams

        # update init params default values form @config decorator
        if hasattr(self, "_name"):
            # we need to initialise config
            # get init params set by @config decorator
            default_init_params = getattr(
                self.component_class, "_jembe_config_init_params", dict()
            )
            init_params = default_init_params.copy()
            init_params.update(kwargs.copy())
            init_parameters = signature(init_method).parameters

            init_named_params = [
                name
                for name, param in init_parameters.items()
                if param.kind == Parameter.POSITIONAL_OR_KEYWORD and name != "self"
            ]
            init_has_kwarg_param = (
                len(
                    [
                        name
                        for name, param in init_parameters.items()
                        if param.kind == Parameter.VAR_KEYWORD
                    ]
                )
                > 0
            )
            try:
                if init_has_kwarg_param:
                    # init has **kwargs param so use all init_params suplied
                    init_method(self, *args, **init_params)
                else:
                    # from init-params use only params that actual exesit in __init__
                    # becouse __init__ does not have **kwargs
                    init_method(
                        self,
                        *args,
                        **{
                            key: value
                            for key, value in init_params.items()
                            if key in init_named_params[len(args) :]
                        },
                    )
            except Exception as e:
                print("Error in {kwargs} __init__: {e}")
                raise e
        else:
            # Component config is used inside @config or @page decorator
            # no need to proper initialise class
            # just set _raw_init_params that will be picked up by @config or @page
            # decorator and used to set _jembe_config_init_params
            self._raw_init_params = kwargs.copy()

    return decoratedInit


class ComponentConfigMeta(ABCMeta):
    def __new__(cls, name, bases, attrs, **kwargs):
        # decorate __init__ to create _raw_init_params dict
        if "__init__" in attrs:
            attrs["__init__"] = componentConfigInitDecorator(attrs["__init__"])
        new_class = super().__new__(cls, name, bases, dict(attrs), **kwargs)
        return new_class


class ComponentConfig(metaclass=ComponentConfigMeta):
    """
    Compononent config defines behavior of all instances of component that
    are known at build time, like: url_path, subcomponents, name etc.
    """

    # initialised by _jembe_init
    _name: str
    _component_class: Type["jembe.Component"]
    _parent: Optional["jembe.ComponentConfig"]

    # initilised by _jembe_prepare_component_init run inside _jembe_init
    component_actions: Dict[str, "ComponentAction"]  # [method_name]
    component_listeners: Dict[str, "ComponentListener"]  # [method_name]
    redisplay: Tuple["jembe.RedisplayFlag", ...]
    _hiearchy_level: int
    _url_params: Tuple["UrlParamDef", ...]
    _key_url_param: "UrlParamDef"

    # intialise by Jembe app after registring route
    __endpoint: str

    @classmethod
    def _jembe_init_(
        cls,
        _name: str,
        _component_class: Type["jembe.Component"],
        _parent: Optional["jembe.ComponentConfig"],
        **init_params,
    ):
        """
        Instance creation by explicitly calling __new__ and __init__
        becouse _parent should be avaible in __init__
        """
        cconfig = object.__new__(cls)
        cconfig._name = _name
        cconfig._component_class = _component_class
        cconfig._parent = _parent
        cconfig._jembe_prepare_component_init()
        cconfig.__init__(**init_params)  # type:ignore
        return cconfig

    def _jembe_prepare_component_init(self):
        """
        Called by _jembe_init_, when actually inicitialising component
        to calculate and set all attributes required by jembe framework.
        Reads component class description and sets appropriate config params
        like url_path state and init params, etc.

        This method is not run when you initiate Config inside @config or @page
        decorator in order to set default values.
        """
        ### attributes configured by component_class
        # obtain component actions
        display_method = getattr(self.component_class, self.DEFAULT_DISPLAY_ACTION)
        self.component_actions = {
            self.DEFAULT_DISPLAY_ACTION: ComponentAction.from_method(display_method)
        }
        for method_name, method in getmembers(
            self.component_class,
            lambda o: isfunction(o) and getattr(o, "_jembe_action", False),
        ):
            self.component_actions[method_name] = ComponentAction.from_method(method)

        # obtain component listeners
        self.component_listeners = dict()
        for method_name, method in getmembers(
            self.component_class,
            lambda o: isfunction(o) and getattr(o, "_jembe_listener", False),
        ):
            self.component_listeners[method_name] = ComponentListener.from_method(
                method
            )

        # set redisplay from @redisplay decorator or with default value
        redisplay_settings = getattr(display_method, "_jembe_redisplay", ())
        if redisplay_settings:
            self.redisplay = redisplay_settings
        else:
            self.redisplay = self.REDISPLAY_DEFAULT_FLAGS

        ### attributes configured by parent
        self._hiearchy_level = 0 if not self.parent else self.parent._hiearchy_level + 1
        # Gets all component.__init__ paramters without default value
        # and populate self._url_params
        self._url_params = tuple(
            UrlParamDef.from_inspect_parameter(p, self._hiearchy_level)
            for p in self.component_class._jembe_init_signature.parameters.values()
            if p.default == p.empty and p.name != "self" and not p.name.startswith("_")
        )
        self._key_url_param = UrlParamDef(
            self.KEY_URL_PARAM_NAME,
            identifier=calc_url_param_identifier(
                self.KEY_URL_PARAM_NAME, self._hiearchy_level
            ),
            convertor=UrlConvertor.STR0,
        )

    KEY_URL_PARAM_NAME = "component_key"
    KEY_URL_PARAM_SEPARATOR = "."
    DEFAULT_DISPLAY_ACTION = "display"
    DEFAULT_AC_EXCEPTION = AccessDenied

    _raw_init_params: dict

    WHEN_STATE_CHANGED = RedisplayFlag.WHEN_STATE_CHANGED
    WHEN_DISPLAY_EXECUTED = RedisplayFlag.WHEN_DISPLAY_EXECUTED
    WHEN_ON_PAGE = RedisplayFlag.WHEN_ON_PAGE

    REDISPLAY_DEFAULT_FLAGS = (RedisplayFlag.WHEN_STATE_CHANGED,)

    def __init__(
        self,
        template: Optional[Union[str, Iterable[str]]] = None,
        components: Optional[Dict[str, "jembe.ComponentRef"]] = None,
        inject_into_components: Optional[
            Callable[["jembe.Component", "jembe.ComponentConfig"], dict]
        ] = None,  # TODO
        redisplay: Tuple["jembe.RedisplayFlag", ...] = (),
        changes_url: bool = True,
        url_query_params: Optional[Dict[str, str]] = None,
    ):
        """_summary_

        template: path to default template for displaying component
        components: Definition of subcomponents
        inject_into_components: Callable to inject init params into subcomponents
        redisplay: Flag to define when componnet will be redisplayed on client depending of its state
        changes_url: Does this component changes location url and allow back browser navigation to it
        url_query_params: Mapping from GET Query params to state params used when component is called directly via regular http get request dict(<name of get queryparam> = <name of state param>)
        Args:
            template (Optional[Union[str, Iterable[str]]], optional): _description_. Defaults to None.
            components (Optional[Dict[str, &quot;jembe.ComponentRef&quot;]], optional): _description_. Defaults to None.
            inject_into_components (Optional[ Callable[[&quot;jembe.Component&quot;, &quot;jembe.ComponentConfig&quot;], dict] ], optional): _description_. Defaults to None.
            changes_url (bool, optional): _description_. Defaults to True.
            url_query_params (Optional[Dict[str, str]], optional): _description_. Defaults to None.
        """

        # use default template if template name is not provided
        if template is None:
            self.template: Tuple[str, ...] = self.default_template_names
        elif isinstance(template, str):
            self.template = (template,)
        else:
            templates: List[str] = []
            for t in template:
                if t == "":
                    templates.extend(self.default_template_names)
                else:
                    templates.append(t)
            self.template = tuple(templates)

        self.components: Dict[str, "jembe.ComponentRef"] = (
            components if components else dict()
        )
        self._inject_into_components = inject_into_components

        # if redisplay is set use it, otherwise leave
        # redisplay set by @redisplay decorator or default value
        if redisplay:
            self.redisplay = redisplay
        self.changes_url = changes_url
        self.url_query_params = (
            url_query_params if url_query_params is not None else dict()
        )

        if not self.changes_url:
            # set changes_url to False to all its children components
            self.update_components_config(None, dict(changes_url=False))

    @property
    def endpoint(self) -> str:
        try:
            return self.__endpoint
        except AttributeError:
            raise JembeError(f"Endpoint not set by jembe app: {self.full_name}")

    @endpoint.setter
    def endpoint(self, endpoint: str):
        self.__endpoint = endpoint

    @property
    def component_class(self) -> Type["jembe.Component"]:
        try:
            return getattr(self, "_component_class")
        except AttributeError:
            raise JembeError(
                f"Component Config {self.full_name} is not initialised properly, 'component_class' is missing."
            )

    @property
    def name(self) -> str:
        try:
            return getattr(self, "_name")
        except AttributeError:
            raise JembeError(
                f"Component Config {self.full_name} is not initialised properly, 'name' is missing."
            )

    @property
    def parent(self) -> Optional["jembe.ComponentConfig"]:
        try:
            return getattr(self, "_parent")
        except AttributeError:
            raise JembeError(
                f"Component Config {self.full_name} is not initialised properly, 'parent' is missing."
            )

    @property
    def full_name(self) -> str:
        if self.parent:
            return f"{self.parent.full_name}/{self.name}"
        return f"/{self.name}"

    @property
    def url_path(self) -> str:
        if not self._url_params:
            url_path = f"/{self.name}{self._key_url_param.url_pattern}"
        else:
            url_params = "/".join(up.url_pattern for up in self._url_params)
            url_path = f"/{self.name}{self._key_url_param.url_pattern}/{url_params}"
        url_path = f"{self.parent.url_path}{url_path}" if self.parent else url_path
        return url_path

    @property
    def hiearchy_level(self) -> int:
        try:
            return self._hiearchy_level
        except AttributeError:
            raise JembeError("Componet config initialisation is not complete")

    def get_raw_url_params(self, state: "ComponentState", key: str) -> dict:
        url_params = {up.identifier: state[up.name] for up in self._url_params}
        url_params[self._key_url_param.identifier] = (
            f"{self.KEY_URL_PARAM_SEPARATOR}{key}" if key else ""
        )
        return url_params

    def build_url(
        self, exec_name: str, component: Optional["jembe.Component"] = None
    ) -> str:
        from .app import get_processor

        processor: "jembe.Processor" = get_processor()
        exec_names = tuple(
            accumulate(map(lambda x: "/" + x, exec_name.strip("/").split("/")), add)
        )
        url_params = dict()
        for en in exec_names:
            if component is not None and en == component.exec_name:
                cmp = component
            else:
                cmp = processor.components[en]

            url_params.update(cmp._config.get_raw_url_params(cmp.state, cmp.key))
        return url_for(self.endpoint, **url_params)

    @property
    def default_template_names(self) -> Tuple[str, ...]:
        tname = f"{self.full_name.strip('/')}.html"
        return (tname, f"pages/{tname}")

    def update_components_config(
        self,
        name: Optional[str],
        params: Dict,
    ):
        """
        For use inside __init__ to change sub components config init params
        regardless if component is referenced with or without componentconfig part.

        name: name of the component config to change if name is None apply params to all components
        params: dict with new params to set
        """
        if not self.components:
            return

        def _update_cref(cname: str, params: Dict):
            """Update self.compoennts"""
            comp_ref: "jembe.ComponentRef" = self.components[cname]
            if not isinstance(comp_ref, tuple):
                comp_ref = (comp_ref, dict())
                self.components[cname] = comp_ref
            if isinstance(comp_ref[1], ComponentConfig):
                comp_ref = (comp_ref[0], comp_ref[1]._raw_init_params)
                self.components[cname] = comp_ref
            if isinstance(comp_ref[1], dict):  # just to satisfy typing check
                comp_ref[1].update(params)
            else:
                # this shuld never happend
                raise NotImplementedError()

        if name is None:
            for name in self.components.keys():
                _update_cref(name, params)
        else:
            if name not in self.components:
                raise ValueError(
                    f"Component with name {name} is not defined inside {self.name}"
                )
            _update_cref(name, params)

    @property
    def components_configs(self) -> Dict[str, "jembe.ComponentConfig"]:
        from .app import get_processor

        processor: "jembe.Processor" = get_processor()
        configs: Dict[str, "jembe.ComponentConfig"] = dict()
        for component_name in self.components.keys():
            configs[component_name] = processor.jembe.components_configs[
                f"{self.full_name}/{component_name}"
            ]
        return configs

    @property
    def components_classes(self) -> Dict[str, Type["jembe.Component"]]:
        def _get_cref_class(cref: "jembe.ComponentRef") -> Type["jembe.Component"]:
            cc = cref[0] if isinstance(cref, tuple) else cref
            if isinstance(cc, str):
                return import_by_name(cc)
            return cc

        return {name: _get_cref_class(cref) for name, cref in self.components.items()}

    @property
    def super(self):
        return super(self.__class__, self)



import os
import pytest
from jembe import Jembe
from jembe.files import Storage
from jembe.exceptions import JembeError

from flask import Flask as _Flask


class Flask(_Flask):
    testing = True
    secret_key = "test key"

class TempTestStorage(Storage):
    pass

@pytest.fixture
def app():
    app = Flask("flask_test", root_path=os.path.dirname(__file__))

    @app.errorhandler(JembeError)
    def handle_jembe_errror(error):
        app.logger.exception(error)
        return "JembeError", 500
    yield app


@pytest.fixture
def app_ctx(app):
    with app.app_context() as ctx:
        yield ctx


@pytest.fixture
def req_ctx(app):
    with app.test_request_context(path="/") as ctx:
        yield ctx


@pytest.fixture
def jmb(app):
    yield Jembe(app)


@pytest.fixture
def client(app):
    yield app.test_client()



from jembe.processor import EmitCommand

gmen = EmitCommand._glob_match_exec_name
"""
if pattern is:

    - None -> match to every initialised component 
    - /compoent1.key1/compoent2.key2    -> compoenent with complete exec_name
    - component                       -> match to direct child named "component" without key
    - component.*                     -> match to direct child named "component" with any key
    - component.key                   -> match to direct child named "component with key equals "key"
    - *                               -> match to any direct child
    - **/component[.[*|<key>]]        -> match to child at any level
    - ..                                -> match to parent
    - ../component[.[*|<key>]]          -> match to sibling 
    - /**/.                             -> match to parent at any level
    - /**/component[.[*|<key>]]/**/.    -> match to parent at any level named
    - etc.
"""


def test_match_any_component():
    sen = "/pa/ca/cb"
    pattern = None
    assert gmen(sen, pattern, "/pa/ca/cb")
    assert gmen(sen, pattern, "/pa/ca/cb.k2")
    assert gmen(sen, pattern, "/pa/ca/cb/cc")
    assert gmen(sen, pattern, "/pa/ca")
    assert gmen(sen, pattern, "/pb/cc")


def test_match_by_complete_name():
    sen = "/pa/ca/cb"
    assert gmen(sen, "/pa/ca/cb", "/pa/ca/cb")
    assert gmen(sen, "/pa/ca/cb", "/pa/ca/cb.k2") == False
    assert gmen(sen, "/pa/cc", "/pa/cc")
    assert gmen(sen, "/pa/cc.k1", "/pa/ca") == False
    assert gmen(sen, "/pb/cc.k1", "/pb/cc.k1")


def test_match_direct_chid_complete_name():
    sen = "/pa/ca"
    assert gmen(sen, "cb", "/pa/ca/cb")
    assert gmen(sen, "cb", "/pa/ca/cb.k1") == False
    assert gmen(sen, "cb.k1", "/pa/ca/cb") == False
    assert gmen(sen, "cb.k1", "/pa/ca/cb.k1")
    assert gmen(sen, "cb/cc", "/pa/ca/cb/cc")
    assert gmen(sen, "cb.k1/cc.k2", "/pa/ca/cb.k1/cc.k2")
    assert gmen(sen, "cb/cc", "/pa/ca/cb.k1/cc.k2") == False


def test_match_direct_chid_search_pattern():
    sen = "/pa/ca"
    assert gmen(sen, "cb.*", "/pa/ca/cb") == False
    assert gmen(sen, "cb.*", "/pa/ca/cb.k1") == True
    assert gmen(sen, "cb.*/cc.*", "/pa/ca/cb/cc") == False
    assert gmen(sen, "cb.*/cc.*", "/pa/ca/cb/cc.k2") == False
    assert gmen(sen, "cb.*/cc.*", "/pa/ca/cb.k1/cc.k2") == True
    assert gmen(sen, "cb.*/**/cc.*", "/pa/ca/cb.k1/cc.k2") == True
    assert gmen(sen, "cb.*/**/cc.*", "/pa/ca/cb.k1/cd/cc.k2") == True
    assert gmen(sen, "cb.*/**/cc.*", "/pa/ca/cb.k1/cd.kd/cc.k2") == True


def test_match_go_back_seach_pattern():
    sen = "/pa/ca"
    assert gmen(sen, "..", "/pa")
    assert gmen(sen, "../../pb/cb.*", "/pb/cb.k1")
    assert gmen(sen, "../cb.*/cc.*", "/pa/cb/cc") == False
    assert gmen(sen, "../cb.*/cc.*", "/pa/cb.k1/cc") == False
    assert gmen(sen, "../cb.*/cc.*", "/pa/cb.k1/cc.k2")
    assert gmen(sen, "../**/cb.*/cc.*", "/pa/cb.k1/cc.k2")
    assert gmen(sen, "../**/cb.*/cc.*", "/pa/cd/ck.kk/cb.k1/cc.k2")


def test_match_parrent_seach_pattern():
    sen = "/pa/ca/cb/cc/cd"
    assert gmen(sen, "/**/.", "/pa/ca/cb")
    assert gmen(sen, "/**/.", "/pa/ca")
    assert gmen(sen, "/**/.", "/pa/ca/cb/cc")
    assert gmen(sen, "/**/.", "/pb/ca/cb/cc") == False
    assert gmen(sen, "/**/cb/**/.", "/pa/ca/cb")
    sen = "/pa/ca/cb.k1/cc/cd"
    assert gmen(sen, "/**/cb.k1/**/.", "/pa/ca.ka/cb.k1") == False
    assert gmen(sen, "/**/cb.k1/**/.", "/pa/ca/cb.k1")
    sen = "/pa/ca/cb.k1/cc/cd"
    assert gmen(sen, "/**/cb.*/**/.", "/pa/ca/cb.k1")


def test_match_children():
    assert gmen("/pa/ca", "*", "/pa/ca/cb")
    assert gmen("/pa/ca", "*", "/pa/ca/cb.kb")
    assert gmen("/pa/ca", "*", "/pa/ca/cb.kb/cc") == False
    assert gmen("/pa/ca", "*.*", "/pa/ca/cb") == False
    assert gmen("/pa/ca", "*.*", "/pa/ca/cb.kb")
    assert gmen("/pa/ca", "*.*", "/pa/ca/cb.kb/cc") == False
    assert gmen("/pa/ca", "**/*", "/pa/ca/cb")
    assert gmen("/pa/ca", "**/*", "/pa/ca/cb/cc")
    assert gmen("/pa/ca", "**/*", "/pa/ca/cb.kb/cc")
    assert gmen("/pa/ca", "**/*", "/pa/ca/cb.kb/cc.kc")
    assert gmen("/pa/ca", "**/*.*", "/pa/ca/cb") == False
    assert gmen("/pa/ca", "**/*.*", "/pa/ca/cb/cc") == False
    assert gmen("/pa/ca", "**/*.*", "/pa/ca/cb.kb/cc") == False
    assert gmen("/pa/ca", "**/*.*", "/pa/ca/cb.kb/cc.kc")
    assert gmen("/pa/ca", "**/*.*", "/pa/ca/cb/cc.kc")



from io import BytesIO
from flask import json


def test_upload_files(jmb, client):
    r = client.post(
        "/jembe/upload_files",
        data=dict(
            uploadId1=(BytesIO(b"FILE1 CONTENT"), "file1.txt"),
            uploadId2=(BytesIO(b"FILE2 CONTENT"), "file2.txt"),
        ),
        headers={"x-jembe": "upload"},
    )
    assert r.status_code == 200

    rd = json.loads(r.data)
    assert "fileUploadResponseId" in rd
    assert rd["fileUploadResponseId"] is not None
    assert "files" in rd
    assert len(rd["files"].keys()) == 2
    assert rd["files"]["uploadId1"] == [
        dict(
            path="UPLOADS/{}/file1.txt".format(rd["fileUploadResponseId"]),
            storage="temp",
        )
    ]
    assert rd["files"]["uploadId2"] == [
        dict(
            path="UPLOADS/{}/file2.txt".format(rd["fileUploadResponseId"]),
            storage="temp",
        )
    ]

    r = client.post(
        "/jembe/upload_files",
        data=dict(
            uploadId1=(BytesIO(b"FILE1 CONTENT"), "file1.txt"),
            uploadId2=(BytesIO(b"FILE2 CONTENT"), "file1.txt"),
        ),
        headers={"x-jembe": "upload"},
    )
    assert r.status_code == 200

    rd = json.loads(r.data)
    assert rd["files"]["uploadId1"][0]["path"] != rd["files"]["uploadId2"][0]["path"]



from typing import Dict, Optional, TYPE_CHECKING, List
from jembe.defaults import DEFAULT_TEMP_STORAGE_UPLOAD_FOLDER
from uuid import uuid1
from flask import jsonify
from werkzeug.utils import secure_filename
from .component_config import UrlPath, config
from .component import Component
from .files import File

# from flask import send_from_directory
from .app import get_storage, get_temp_storage, get_processor

if TYPE_CHECKING:
    from .common import DisplayResponse


class DisplayFileComponent(Component):
    """Serve files from Jembe Storages on Direct HTTP request"""

    def __init__(self, storage_name: str, file_path: UrlPath):
        super().__init__()

    def display(self) -> "DisplayResponse":
        return get_storage(self.state.storage_name).send_file(self.state.file_path)


class UploadFilesComponent(Component):
    """
    Saves uploaded files in temporary storage in new dir named uniquly named fileUploadRequestId.
    Returns json
        dict(files = dict(fileUploadId, [{storage=storage_name, path=file_path}]), fileUploadResponseId = fileUplaodResponseId)

    This component is only component that directly interecat with
    http post request in order to obtain files from it.
    It does not uses state parama.

    Basicaly this component create state params of file instances for other
    components so that thay can use file as init/state param.
    """

    def __init__(self):
        self.files: Dict[str, List[File]] = dict()
        self.file_upload_response_id: Optional[str] = None
        super().__init__()

    def _save_files_to_temp_storage(self):
        """Saves files to temp storage and populates self.files and self.file_upload_response_id"""
        # checks if this is valid upload request
        processor = get_processor()
        if processor.is_x_jembe_upload_request:
            # procede with upload
            temp_storage = get_temp_storage()

            # generate unique file_upload_response_id
            while self.file_upload_response_id is None or temp_storage.exists(
                "/".join(
                    [DEFAULT_TEMP_STORAGE_UPLOAD_FOLDER, self.file_upload_response_id]
                )
            ):
                self.file_upload_response_id = secure_filename(str(uuid1()))
            temp_storage.makedirs(
                "/".join(
                    [DEFAULT_TEMP_STORAGE_UPLOAD_FOLDER, self.file_upload_response_id]
                )
            )

            for fileUploadId, file in processor.request.files.items(True):
                if file.filename != "":
                    if file and self.allowed_file(file.filename):
                        tmp_file = temp_storage.store_file(
                            file,
                            "/".join(
                                [
                                    DEFAULT_TEMP_STORAGE_UPLOAD_FOLDER,
                                    self.file_upload_response_id,
                                ]
                            ),
                            True,
                        )
                        if fileUploadId in self.files:
                            self.files[fileUploadId].append(tmp_file)
                        else:
                            self.files[fileUploadId] = [tmp_file]

    def display(self) -> "DisplayResponse":
        self._save_files_to_temp_storage()

        return jsonify(
            dict(
                files={
                    fid: [File.dump_init_param(f) for f in ffs]
                    for fid, ffs in self.files.items()
                },
                fileUploadResponseId=self.file_upload_response_id,
            )
        )

    def allowed_file(self, file_name: str) -> bool:
        # TODO
        return True


@config(
    Component.Config(
        components=dict(file=DisplayFileComponent, upload_files=UploadFilesComponent)
    )
)
class JembePage(Component):
    pass



from werkzeug import exceptions as we


class JembeError(Exception):
    pass


class ComponentPreviousStateUnavaiableError(JembeError):
    pass


class BadRequest(we.BadRequest):
    pass


class Unauthorized(we.Unauthorized):
    pass


class Forbidden(we.Forbidden):
    pass


class NotFound(we.NotFound):
    pass


class AccessDenied(NotFound):
    pass


class Conflict(we.Conflict):
    pass


class Gone(we.Gone):
    pass


class InternalServerError(we.InternalServerError):
    pass


class NotImplemented(we.NotImplemented):
    pass



from jembe.utils import page_url
from typing import TYPE_CHECKING
from jembe import Component, page_url

if TYPE_CHECKING:
    from jembe import Jembe


def test_page_url_on_page(jmb: "Jembe", req_ctx):
    class AComponent(Component):
        pass

    class BComponent(Component):
        def __init__(self, b1: int) -> None:
            super().__init__()

    @jmb.page(
        "simple_page", Component.Config(components=dict(a=AComponent, b=BComponent))
    )
    class SimplePage(Component):
        pass

    with req_ctx:
        assert page_url("simple_page") == "/simple_page"
        assert page_url("simple_page.key1") == "/simple_page.key1"
        assert page_url("simple_page/a.key2") == "/simple_page/a.key2"
        assert page_url("simple_page/b", [{}, {"b1": 1}]) == "/simple_page/b/1"



from typing import Sequence, TYPE_CHECKING, Optional, Tuple, Type, List, Dict, Any
from os import path
from .defaults import (
    DEFAULT_JEMBE_MEDIA_FOLDER,
    PRIVATE_STORAGE_NAME,
    PUBLIC_STORAGE_NAME,
    TEMP_STORAGE_NAME,
)
from flask import Blueprint, request
from .processor import Processor
from .exceptions import JembeError
from flask import g, current_app
from .common import ComponentRef, exec_name_to_full_name, import_by_name


if TYPE_CHECKING:  # pragma: no cover
    import jembe
    from flask import Flask, Response


class _JembeState:
    def __init__(self, jembe: "Jembe") -> None:
        self.jembe = jembe


class Jembe:
    """Tracks all registred Components and Storages of an application.

    ``Jembe`` is initilised only once per application and it references to
    all registred Components and Storages used by the application.

    ``Jembe`` instance is created and associated with Flask application instance
    similar to other Flask extensions.

    .. code-block:: python

        from flask import Flask
        from jembe import Jembe

        app = Flask(__name__)
        jmb = Jembe(app)

    or when creating flask instance dynamicaly:

    .. code-block:: python

        from flask import Flask
        from jembe import Jembe

        jmb = Jembe()

        def create_app(config):
            app = Flask(__name__)
            jmb.init_app(app)
            return app

    Attributes:
        app:
            Optional Flask application instance.

            If Flask application instance is not provided than ``Jembe.init_app``
            method must be executed after Flask instance becomes avaiable.

        storages:
            Optional Jembe storages definitions.

            Files created or uploaded by end user are saved in storages. Jembe instance
            can save and access files only in storages registred here.

            If storages is not provided than default storages configuration will be used:

            - Public storage named `public` in ``data/media/public`` directory;
            - Private storage named `private` in ``data/media/private`` directory;
            - Temporary storage named `temp` in ``data/media/temp`` directory;

            .. code-block:: python

                from jembe import DiskStorage

                # example storage configuration
                Jembe(storages=[
                    DiskStorage("public", "data/media/public"),
                    DiskStorage("private", "data/media/private", type=DiskStorage.Type.PRIVATE),
                    DiskStorage("temp", "data/media/temp", type=DiskStorage.Type.TEMP),
                ])

            Paths in DiskStorage definitions are relateve to ``JEMBE_MEDIA_FOLDER`` Flask configuraration variable.  To change location of ``JEMBE_MEDIA_FOLDER`` alter Flask config variable.

            Example of changing ``JEMBE_MEDIA_FOLDER`` when working on project created from jembe startproject template:

            .. code-block:: python
                :caption: instance/config.py

                JEMBE_MEDIA_FOLDER = "/var/yourproject/media"

            .. warning:: When You manually define storages, you must define all storages used by your application including at least one temporary storage.

    Raises:
        JembeError: More then one Jembe extension is initialised for a Flask instance;
        JembeError: Storage is initialised before associating Jembe with Flask instance;
        JembeError: Temporary Storage is not configured;

    Returns:
        :obj:`Jembe`: Jembe instance
    """

    X_JEMBE = "X-Jembe"
    X_RELATED_UPLOAD = "X-Jembe-Related-Upload"

    def __init__(
        self,
        app: Optional["Flask"] = None,
        storages: Optional[Sequence["jembe.Storage"]] = None,
    ):
        """Initialise jembe configuration"""
        self.__flask: "Flask"
        # all registred jembe components configs [full_name, config instance]
        self.components_configs: Dict[str, "jembe.ComponentConfig"] = {}
        # pages waiting to be registred
        self._unregistred_pages: Dict[str, "jembe.ComponentRef"] = {}

        self._storages: Dict[str, "jembe.Storage"]
        self.extensions: Dict[str, Any] = dict()
        self.initialised_extensions: List[str] = []

        from .jembe_page import JembePage

        self.add_page("jembe", JembePage)

        if app is not None:
            self.init_app(app, storages)

    @property
    def flask(self) -> Optional["Flask"]:
        """Returns associated Flask instance"""
        try:
            return self.__flask
        except AttributeError:
            return None

    def init_app(
        self, app: "Flask", storages: Optional[Sequence["jembe.Storage"]] = None
    ):
        """
        Use this callback to initialise Jembe with Flask application when Flask
        application is created dynamicaly.
        """
        self.__flask = app

        # init extensions
        for k, w in self.extensions.items():
            if k not in self.initialised_extensions:
                w.do_init_jembe()
                self.initialised_extensions.append(k)

        # app.teardown_appcontext(self.teardown)
        # app.context_processor(self.template_processor)
        if "jembe" in self.__flask.extensions:
            raise JembeError(
                "Only one Jembe extension can be initialise for a Flask instance."
            )
        self.__flask.extensions["jembe"] = _JembeState(self)

        # Init storages
        self._init_storages(storages)

        # register all unregistred pages added to app before associating
        # app with flask instance
        if self._unregistred_pages:
            for name, component in self._unregistred_pages.items():
                self._register_page(name, component)
            self._unregistred_pages = {}

    def _init_storages(self, storages: Optional[Sequence["jembe.Storage"]]):
        if not self.flask:
            raise JembeError(
                "Cannot initialise storages before "
                "initialising jembe with flask instance"
            )
        from .files import DiskStorage

        if storages is None and not hasattr(self, "_storages"):
            # initialise default storages

            media_folder = self.flask.config.get(
                "JEMBE_MEDIA_FOLDER", DEFAULT_JEMBE_MEDIA_FOLDER
            )
            storages = [
                DiskStorage(PUBLIC_STORAGE_NAME, path.join(media_folder, "public")),
                DiskStorage(
                    PRIVATE_STORAGE_NAME,
                    path.join(media_folder, "private"),
                    type=DiskStorage.Type.PRIVATE,
                ),
                DiskStorage(
                    TEMP_STORAGE_NAME,
                    path.join(media_folder, "temp"),
                    type=DiskStorage.Type.TEMP,
                ),
            ]
        if storages is None or not next(
            (s for s in storages if s.type == DiskStorage.Type.TEMP), False
        ):
            raise JembeError(
                "Temporary Storage must be configured in order for file upload to work"
            )

        self._storages = {s.name: s for s in storages}

    def add_page(
        self,
        name: str,
        component: Type["jembe.Component"],
        component_config: Optional["jembe.ComponentConfig"] = None,
    ):
        """
        Register Component as "Page Component" to Jembe instance.

        Page Component is root component in hiearchy.
        One application can have multiple Page Components.

        Page Component and all its sub-component will be recursivlly
        registred.
        Page component becomes avaiable under `/<name>` URL.

        Args:
            name: Unique name of the component
            component: Component Class
            component_config: Optional instance of Component.Config
                configuring Component behavior
        """
        component_ref: ComponentRef = (
            (
                component,
                component_config,
            )
            if component_config
            else component
        )

        if self.flask is not None:
            self._register_page(name, component_ref)
        else:
            self._unregistred_pages[name] = component_ref

    def page(
        self,
        name: str,
        component_config: Optional["jembe.ComponentConfig"] = None,
    ):
        """
        Decorator that registers Component as "Page Component" to Jembe instance.

        It does same thing as ``add_page`` but as decorator:

        .. code-block:: python

            from jembe import Component
            from my_project import jmb

            @jmb.page("simple", Component.Config(components={..}))
            class SimplePage(Component):
                pass
        """

        def decorator(component):
            self.add_page(name, component, component_config)
            return component

        return decorator

    def _register_page(self, name: str, component_ref: ComponentRef):
        if self.flask is None:  # pragma: no cover
            raise NotImplementedError()

        # go down component hiearchy
        bp: Optional["Blueprint"] = None
        component_refs: List[
            Tuple[str, ComponentRef, Optional["jembe.ComponentConfig"]]
        ] = [(name, component_ref, None)]
        while component_refs:
            component_name, curent_ref, parent_config = component_refs.pop(0)
            if isinstance(curent_ref, tuple):
                # create config with custom params
                component_class: Type["jembe.Component"] = (
                    curent_ref[0]
                    if not isinstance(curent_ref[0], str)
                    else import_by_name(curent_ref[0])
                )
                curent_ref_params = (
                    curent_ref[1]
                    if isinstance(curent_ref[1], dict)
                    else curent_ref[1]._raw_init_params
                )
                component_config = component_class.Config._jembe_init_(
                    _name=component_name,
                    _parent=parent_config,
                    _component_class=component_class,
                    **curent_ref_params,
                )
            else:
                # create config with default params
                component_class = (
                    curent_ref
                    if not isinstance(curent_ref, str)
                    else import_by_name(curent_ref)
                )
                component_config = component_class.Config._jembe_init_(
                    _name=component_name,
                    _parent=parent_config,
                    _component_class=component_class,
                )
            # fill components_configs
            self.components_configs[component_config.full_name] = component_config

            if bp is None:
                bp = Blueprint(
                    component_name,
                    component_class.__module__,
                    template_folder="templates",
                    static_folder="static",
                    static_url_path=f"/{component_name}/static",
                )

            bp.add_url_rule(
                component_config.url_path,
                component_config.full_name,
                jembe_master_view,
                methods=["GET", "POST"],
            )
            component_config.endpoint = f"{bp.name}.{component_config.full_name}"

            if component_config.components:
                component_refs.extend(
                    (name, cref, component_config)
                    for name, cref in component_config.components.items()
                )

        if bp:
            self.flask.register_blueprint(bp)

    def get_component_config(self, exec_name: str) -> "jembe.ComponentConfig":
        try:
            return self.components_configs[exec_name_to_full_name(exec_name)]
        except KeyError:
            raise JembeError(
                f"Component {exec_name_to_full_name(exec_name)} does not exist"
            )

    def get_storage_by_type(
        self, storage_type: "jembe.Storage.Type", storage_name: Optional[str] = None
    ) -> "jembe.Storage":
        """
        Returns first storage of the provided type if it exists.

        Args:
            storage_type: Storage type one of `Storage.Type.PRIVATE`, `Storage.Type.PUBLIC` or `Storage.Type.TEMP`
            storage_name: Optional name of the storage
        Raises:
            JembeError: When storage does not exists
        """
        # returs named storage if exist and it's right type
        if storage_name is not None:
            try:
                if self._storages[storage_name].type == storage_type:
                    return self._storages[storage_name]
                else:
                    raise JembeError(
                        f"Storage '{storage_name}' is not '{storage_type.value}'"
                    )
            except KeyError:
                raise JembeError("Storage '{storage_name}' does not exist")

        # returns first storage of adequate type
        try:
            return next(s for s in self._storages.values() if s.type == storage_type)
        except StopIteration:
            raise JembeError(
                f"Storage of type '{storage_type.value}' does not exist"
            )

    def get_storages(self) -> List["jembe.Storage"]:
        """Returns list of all configured Jembe storages"""
        return list(self._storages.values())

    def get_storage(self, storage_name: str) -> "jembe.Storage":
        """Returns storage by storage name

        Args:
            storage_name (str): Name of the storage

        Raises:
            JembeError: Storage does not exist
        """
        try:
            return self._storages[storage_name]
        except KeyError:
            raise JembeError(f"Storage '{storage_name}' does not exist")


def get_processor():
    """Returns current Jembe Processors

    Raises:
        JembeError: When current HTTP request can't be handled by jembe processor or when Jembe extension is not initialised
    """
    if "jmb_processor" not in g:
        if not (request.endpoint and request.blueprint):
            raise JembeError("Request {} can't be handled by jembe processor")
        component_full_name = request.endpoint[len(request.blueprint) + 1 :]
        jembe_state = current_app.extensions.get("jembe", None)
        if jembe_state is None:
            raise JembeError("Jembe extension is not initialised")
        return Processor(jembe_state.jembe, component_full_name, request)
    return g.jmb_processor


def get_jembe() -> "Jembe":
    """Returns Jembe instance processing request for current Flask application

    Raises:
        JembeError: When Jembe extension is not initialised
    """

    jembe_state = current_app.extensions.get("jembe", None)
    if jembe_state is None:
        raise JembeError("Jembe extension is not initialised")
    return jembe_state.jembe


def get_storage(storage_name: str) -> "jembe.Storage":
    """Returns storage by storage name from current Jembe instance

    Args:
        storage_name (str): Name of the storage

    Raises:
        JembeError: Storage does not exist
    """
    try:
        return get_processor().jembe.get_storage(storage_name)
    except JembeError:
        # cant get processor becouse we are not in valid requst context
        # return storage of default jembe app
        return get_jembe().get_storage(storage_name)


def get_storages() -> List["jembe.Storage"]:
    """Returns list of all registred Storages in current Jembe instance"""
    return get_processor().jembe.get_storages()


def get_temp_storage(storage_name: Optional[str] = None) -> "jembe.Storage":
    """Returns temporary storage of current Jembe instance.

    When storage_name is provided it will return temporary storage with given name,
    otherwise it will return first temporary storage from storage list.

    Args:
        storage_name (Optional[str], optional): Name of the temporary storage. Defaults to None.
    """
    from .files import Storage

    return get_processor().jembe.get_storage_by_type(Storage.Type.TEMP, storage_name)


def get_public_storage(storage_name: Optional[str] = None) -> "jembe.Storage":
    """Returns public storage of current Jembe instance.

    When storage_name is provided it will return public storage with given name,
    otherwise it will return first public storage from storage list.

    Args:
        storage_name (Optional[str], optional): Name of the public storage. Defaults to None.
    """
    from .files import Storage

    return get_processor().jembe.get_storage_by_type(Storage.Type.PUBLIC, storage_name)


def get_private_storage(storage_name: Optional[str] = None) -> "jembe.Storage":
    """Returns private storage of current Jembe instance.

    When storage_name is provided it will return private storage with given name,
    otherwise it will return first private storage from storage list.

    Args:
        storage_name (Optional[str], optional): Name of the private storage. Defaults to None.
    """
    from .files import Storage

    return get_processor().jembe.get_storage_by_type(Storage.Type.PRIVATE, storage_name)


def jembe_master_view(**kwargs) -> "Response":
    """Process HTTP request with Jembe Processors"""
    return get_processor().process_request().build_response()



from typing import Any, Dict, List, TYPE_CHECKING, Dict, Optional, Union
from functools import cached_property
from dataclasses import dataclass, field

from flask import json
from jembe.processor import Event
from jembe.exceptions import BadRequest, JembeError
from jembe.component_config import ComponentConfig, redisplay
from jembe import (
    Component,
    action,
    listener,
    config,
    NotFound,
    Unauthorized,
)

if TYPE_CHECKING:
    from jembe import DisplayResponse

    """build simple task apps for testing server side processing and jembe api"""

# data definition
@dataclass
class User:
    username: str


@dataclass
class Task:
    id: int
    title: str
    description: Optional[str] = None
    parent_id: Optional[int] = None


class WipDb:
    # temporary storage of modified and new tasks (Work in Progress)
    def __init__(self, *tasks: Task) -> None:
        self._tasks: Dict[int, Task] = {t.id: t for t in tasks}

    def save(self, original_id: Optional[int] = None) -> Optional[int]:
        map_new_ids: Dict[int, int] = dict()
        tasks = list(self._tasks.values())

        index = 0
        while tasks:
            task = tasks[index]
            if task is None:
                del tasks[index]
                del tasks_db[index]
            elif (
                task.parent_id is None
                or task.parent_id > 0
                or task.parent_id in map_new_ids.keys()
            ):
                # save task
                if task.id < 0:
                    old_id = task.id
                    task.id = (
                        max(tid for tid in tasks_db.keys()) + 1
                        if tasks_db.keys()
                        else 1
                    )
                    map_new_ids[old_id] = task.id

                if task.parent_id and task.parent_id < 0:
                    task.parent_id = map_new_ids[task.parent_id]

                tasks_db[task.id] = task
                del tasks[index]

            if index >= len(tasks):
                index = 0

        if original_id:
            if original_id > 0:
                return original_id
            else:
                return map_new_ids[original_id]
        return None

    def put(self, task: Task):
        self._tasks[task.id] = task

    def get(self, task_id: int) -> Task:
        try:
            return self._tasks[task_id]
        except KeyError:
            raise ValueError("Requested task does not exist")

    def has(self, task_id: int) -> bool:
        return task_id in self._tasks

    @property
    def tasks(self) -> Dict[int, Task]:
        return self._tasks

    def ids(self) -> List[int]:
        return list(self._tasks.keys())


# database
tasks_db: Dict[int, Task] = dict()
session: dict = dict(user=None, wipdbs=dict())

# form
@dataclass
class TaskForm:
    title: str
    description: Optional[str]
    error: Optional[str] = field(default=None, init=False)

    def is_valid(self) -> bool:
        if not self.title:
            self.error = "Title is required"
            return False
        return True

    @classmethod
    def load_init_param(cls, param_value) -> Optional["TaskForm"]:
        return (
            cls(
                title=param_value.get("title", ""),
                description=param_value.get("description", None),
            )
            if param_value is not None
            else None
        )


# components
@config(
    Component.Config(
        # components=dict(
        #     view="tests.test_capp.ViewTask",
        #     edit="tests.test_capp.EditTask",
        #     add="tests.test_capp.AddTask",
        #     delete="tests.test_capp.DeleteTask",
        # ),
        inject_into_components=lambda self, component_config: TaskList.inject_into_components(
            self, component_config
        )
    )
)
class TaskList(Component):
    def inject(self) -> Dict[str, Any]:
        return dict(user=session.get("user", None),)

    @classmethod
    def inject_into_components(cls, self: Component, component_config: ComponentConfig):
        result = dict()
        if self.state.parent_task_id is not None:
            result["parent_task_id"] = self.state.parent_task_id
        if self.state.wip_id is not None:
            result["wip_id"] = self.state.wip_id
        return result

    def __init__(
        self,
        mode: str = "list",
        parent_task_id: Optional[int] = None,
        wip_id: Optional[int] = None,
        user: Optional[User] = None,
    ) -> None:
        if user is None:
            raise Unauthorized()

        self._wipdb = session["wipdbs"][wip_id] if wip_id else None
        if parent_task_id is not None and not (
            parent_task_id in tasks_db
            or (self._wipdb is not None and self._wipdb.has(parent_task_id))
        ):
            raise NotFound()

        if mode != "list" and mode not in self._config.components.keys():
            raise BadRequest()

        self.goto_task_id: Optional[int] = None

        super().__init__()

    @property
    def tasks(self) -> Dict[int, Task]:
        """
        Returns list of tasks where parent_id == state.parent_task_id from 
        tasks_db and wip_db.
        Replacing tasks from tasks_db with task from wip_db if taks have same id
        "new" tasks in wip_db that does not exist in tasks_db must have negative id
        """
        try:
            return self._tasks
        except AttributeError:
            self._tasks: Dict[int, Task] = {
                tid: t
                for tid, t in tasks_db.items()
                if t.parent_id == self.state.parent_task_id
            }
            # replace tasks in db with task from wip
            if self._wipdb:
                for tid, t in self._wipdb.tasks.items():
                    if tid > 0 and tid not in self._tasks:
                        raise ValueError(
                            "Work in progress task that does not exist in db must have negative id"
                        )
                    if t.parent_id == self.state.parent_task_id:
                        self._tasks[tid] = t
            return self._tasks

    @listener(event="_display", source="*")
    def on_display_child(self, event: Event):
        self.state.mode = event.source_name
        self.goto_task_id = event.params.get("task_id", None)

    @listener(event="save", source=["edit", "add"])
    def on_tasks_changed(self, event: Event):
        self.state.mode = "view" if "view" in self._config.components else "edit"
        self.goto_task_id = event.params.get("task_id", None)

    @listener(event="cancel", source="*")
    def on_cancel_operation(self, event: Event):
        self.state.mode = "list"

    @listener(event="delete", source=["delete"])
    def on_delete_task(self, event: Event):
        self.state.mode = "list"

    # redisplay whenever it is executed regardles if the state is changed
    # becouse state changes in task_db or wipdb will not be reflected to the state
    @redisplay(when_executed=True)
    def display(self) -> "DisplayResponse":
        list_template = (
            """<div>"""
            """{% if component("add").is_accessible %}"""
            """<button type="button" jmb-on:click="{{component().jrl}}">Add</button>{% endif %}"""
            "<table>"
            "<tr><th>Task</th><th>Actions</th></tr>"
            "{% for t in tasks.values() %}<tr>"
            "<td>"
            """{% if component("view", task_id=t.id).is_accessible %}"""
            """<a href="{{component().url}}" jmb-on:click="{{component().jrl}}">{{t.title}}</a>"""
            """{% elif component("edit", task_id=t.id).is_accessible %}"""
            """<a href="{{component().url}}" jmb-on:click="{{component().jrl}}">{{t.title}}</a>"""
            # """<a href="{{component.url}}" jmb-on:click="$jmb.component('../view', {task_id:t.id}).display">{{t.title}}</a>"""
            """{% else %}{{t.title}}{% endif %}"""
            "</td>"
            "<td>"
            """{% if component("edit", task_id=t.id).is_accessible %}"""
            """<a href="{{component().url}}" jmb-on:click="{{component().jrl}}">edit</a>{% endif %}"""
            """{% if component("delete", task_id=t.id).is_accessible %}"""
            """<a href="{{component().url}}" jmb-on:click="{{component().jrl}}">delete</a>{% endif %}"""
            "</td>"
            "</tr>{% endfor %}"
            "</table>"
            "</div>"
        )
        if self.state.mode == "list":
            if self.state.parent_task_id is None:
                self.emit("set_page_title", title="Tasks")
            return self.render_template_string(list_template)
        elif self.state.mode == "view":
            if self.goto_task_id:
                return self.render_template_string(
                    "{{component('view', task_id=goto_task_id)}}"
                )
            else:
                return self.render_template_string("{{component('view')}}")
        elif self.state.mode == "edit":
            if self.goto_task_id:
                return self.render_template_string(
                    "{{component('edit', task_id=goto_task_id)}}"
                )
            else:
                return self.render_template_string("{{component('edit')}}")
            return self.render_template_string("{{component('edit')}}")
        elif self.state.mode == "add":
            return self.render_template_string("{{component('add')}}")
        elif self.state.mode == "delete":
            return self.render_template_string(
                "{{component('delete')}}" + list_template
            )
        raise JembeError("invalid mode {}".format(self.state.mode))
        # About strong cupling between compoentes by usint component("../view") which
        # requre that parent component have subcompoenet named view with excepted behavior:
        # This can be avioded by:
        #  1. defining list of components/actions either in init params or in _config that will be
        #     set by parent component
        #  2. setting list of events by parent via config or in init params that will be emited
        #  3. defining list of components/actions factories by parent config etc.
        # in eigher case task list must distinguish components/actions/events that are called/emited on
        # indipendent of current record (add), in context of one recorord (view, edit) or in context of
        # multiple records (delete all)

        # action(title, help_text, icon??, url, jrl,  **whatever)
        # component.action_factory(component_relative_name, title, help_text, icon, init_params, **whatever)


class ViewTask(Component):
    def inject(self) -> Dict[str, Any]:
        return dict(user=session.get("user", None))

    def __init__(
        self, task_id: int, wip_id: Optional[int] = None, user: Optional[User] = None,
    ) -> None:
        if user is None:
            raise Unauthorized()

        self._wipdb = session["wipdbs"][wip_id] if wip_id else None

        if not (
            self.state.task_id in tasks_db
            or (self._wipdb is not None and self._wipdb.has(self.state.task_id))
        ):
            raise NotFound()

        super().__init__()

    @cached_property
    def task(self) -> Task:
        if self._wipdb and self._wipdb.has(self.state.task_id):
            return self._wipdb.get(self.state.task_id)
        else:
            return tasks_db[self.state.task_id]

    def display(self) -> "DisplayResponse":
        self.emit("set_page_title", title="View {}".format(self.task.title))
        return self.render_template_string(
            """<h1><a href="#" jmb-on:click="$jmb.component('..').display()">Back</a> {{task.title}}</h1>"""
            "<div>{{task.description}}</div>"
            "{% if component('subtasks', parent_task_id=task_id).is_accessible %}"
            "<h2>Sub tasks</h2>"
            "<div>{{component()}}</div>"
            "{% endif %}"
        )


class EditTask(Component):
    def inject(self) -> Dict[str, Any]:
        return dict(user=session.get("user", None))

    def __init__(
        self,
        task_id: int,
        form: Optional[TaskForm] = None,
        wip_id: Optional[int] = None,
        user: Optional[User] = None,
    ) -> None:
        if user is None:
            raise Unauthorized()

        self._wipdb = session["wipdbs"][wip_id] if wip_id is not None else None
        if not (
            self.state.task_id in tasks_db
            or (self._wipdb is not None and self._wipdb.has(self.state.task_id))
        ):
            raise NotFound()
        self.__mounted = False

        super().__init__()

    @classmethod
    def load_init_param(
        cls, config: "ComponentConfig", param_name: str, param_value: Any
    ) -> Any:
        if param_name == "form":
            return TaskForm.load_init_param(param_value)
        return super().load_init_param(config, param_name, param_value)

    def mount(self):
        if self.__mounted:
            return

        if self.state.wip_id is None:
            # Initialise wipdb with wip_id
            self.state.wip_id = (
                max(session["wipdbs"].keys()) + 1 if session["wipdbs"] else 1
            )
            self._wipdb = WipDb()
            session["wipdbs"][self.state.wip_id] = self._wipdb

        if self.state.form is None:
            self.state.form = TaskForm(
                title=self.task.title, description=self.task.description
            )

    @cached_property
    def task(self) -> Task:
        if self._wipdb.has(self.state.task_id):
            return self._wipdb.get(self.state.task_id)
        else:
            return tasks_db[self.state.task_id]

    @listener(event=["cancel", "save"], source=".")
    def on_self_save_or_cancel(self, event: Event):
        if not self.isinjected("wip_id"):
            del session["wipdbs"][self.state.wip_id]
            self.state.wip_id = None
        return False

    @action
    def save(self):
        self.mount()
        if self.state.form.is_valid():
            # create new task
            task = Task(
                self.state.task_id,
                self.state.form.title,
                self.state.form.description,
                self.task.parent_id,
            )
            # save in wip_db
            # becaouse we dont want jet to change task in task_db (self.task)
            # do not care if already exist or not
            self._wipdb.put(task)
            if self.isinjected("wip_id"):
                # wipdb is changed
                pass
            else:
                # saving from component who created wip_db
                # save all changes from wip_db
                self._wipdb.save()
            # emit save so that parent can decide what to display next
            self.emit("save", task_id=self.state.task_id)
            # dont redisplay this component after successfull save
            return False
        # form is not valid redisplay it with error

    def display(self) -> "DisplayResponse":
        self.mount()
        self.emit("set_page_title", title="Edit {}".format(self.task.title))
        return self.render_template_string(
            "<div>"
            "<h1>Edit {{task.title}}</h1>"
            """{% if form.error %}<div>{{form.error}}</div>{% endif %}"""
            """<label>Title:"""
            """<input type="text" value="{{form.title}}" """
            """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
            """<label>Description:"""
            """<input type="text" value="{{form.description}}" """
            """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
            """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
            """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
            "{% if component('subtasks', parent_task_id=task_id).is_accessible %}"
            "<h2>Subtasks</h2>"
            "<div>{{component()}}</div>"
            "{% endif %}"
            "</div>"
        )


class AddTask(Component):
    def inject(self) -> Dict[str, Any]:
        return dict(user=session.get("user", None))

    def __init__(
        self,
        parent_task_id: Optional[int] = None,
        task_id: Optional[int] = None,
        form: Optional[TaskForm] = None,
        wip_id: Optional[int] = None,
        user: Optional[User] = None,
    ) -> None:
        if user is None:
            raise Unauthorized()

        self._wipdb = session["wipdbs"][wip_id] if wip_id else None
        if parent_task_id is not None and not (
            parent_task_id in tasks_db
            or (self._wipdb is not None and self._wipdb.has(parent_task_id))
        ):
            raise NotFound()

        self._mounted = False
        super().__init__()

    @classmethod
    def load_init_param(
        cls, config: "ComponentConfig", param_name: str, param_value: Any
    ) -> Any:
        if param_name == "form":
            return TaskForm.load_init_param(param_value)
        return super().load_init_param(config, param_name, param_value)

    def mount(self) -> None:
        if self._mounted:
            return
        self._mounted = True

        if self.state.wip_id is None:
            self.state.wip_id = (
                max(session["wipdbs"].keys()) + 1 if session["wipdbs"] else 1
            )
            self._wipdb = WipDb()
            session["wipdbs"][self.state.wip_id] = self._wipdb

        if self.state.task_id is None:
            self.state.task_id = min([0, *self._wipdb.ids()]) - 1
            new_task = Task(
                id=self.state.task_id, title="", parent_id=self.state.parent_task_id
            )
            self._wipdb.put(new_task)

        if self.state.form is None:
            self.state.form = TaskForm(
                title=self.task.title, description=self.task.description
            )

    @cached_property
    def task(self) -> Task:
        if self._wipdb.has(self.state.task_id):
            return self._wipdb.get(self.state.task_id)
        else:
            return tasks_db[self.state.task_id]

    @listener(event=["cancel", "save"], source=".")
    def on_self_save_or_cancel(self, event: Event):
        if not self.isinjected("wip_id"):
            del session["wipdbs"][self.state.wip_id]
            self.state.wip_id = None
        return False

    @action
    def save(self):
        self.mount()
        if self.state.form.is_valid():
            task = Task(
                self.state.task_id,
                self.state.form.title,
                self.state.form.description,
                self.state.parent_task_id,
            )
            # save in wip_db
            # becaouse we dont want jet to change task in task_db (self.task)
            # do not care if already exist or not
            self._wipdb.put(task)
            if self.isinjected("wip_id"):
                # wipdb is changed
                pass
            else:
                # saving from component who created wip_db
                # save all changes from wip_db
                self.state.task_id = self._wipdb.save(self.state.task_id)
                # remove wipdb becouse we dont need it anymore
                # del session["wipdbs"][self.state.wip_id]
                # self.state.wip_id = None
                # # TODO change injected into of child components when state is changed
            # emit save so that parent can decide what to display next
            self.emit("save", task_id=self.state.task_id)
            # dont redisplay this component after successfull save
            return False
        # form is not valid redisplay it with error

    def display(self) -> "DisplayResponse":
        self.mount()
        self.emit("set_page_title", title="Add task")
        return self.render_template_string(
            "<div>"
            "<h1>New task</h1>"
            """{% if form.error %}<div>{{form.error}}</div>{% endif %}"""
            """<label>Title:"""
            """<input type="text" value="{{form.title}}" """
            """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
            """<label>Description:"""
            """<input type="text" value="{{form.description|default("",true)}}" """
            """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
            """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
            """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
            "{% if component('subtasks').is_accessible %}"
            "<h2>Subtasks</h2>"
            "<div>{{component()}}</div>"
            "{% endif %}"
            "</div>"
        )


class DeleteTask(Component):
    def inject(self) -> Dict[str, Any]:
        return dict(user=session.get("user", None))

    def __init__(
        self, task_id: int, wip_id: Optional[int] = None, user: Optional[User] = None,
    ) -> None:
        if user is None:
            raise Unauthorized()

        self._wipdb = session["wipdbs"][wip_id] if wip_id else None

        if not (
            self.state.task_id in tasks_db
            or (self._wipdb is not None and self._wipdb.has(self.state.task_id))
        ):
            raise NotFound()

        super().__init__()

    @cached_property
    def task(self) -> Task:
        if self._wipdb and self._wipdb.has(self.state.task_id):
            return self._wipdb.get(self.state.task_id)
        else:
            return tasks_db[self.state.task_id]

    def delete(self):
        if (
            self.isinjected("wip_id")
            and self._wipdb
            and self._wipdb.has(self.state.task_id)
        ):
            # delete from wip_db
            self.state.wip_db[self.state.task_id] = None
            self.emit("delete")
        else:
            # direct delte from tasks_db
            del tasks_db[self.state.task_id]
            self.emit("delete")
        # dont redisplay parent (Tasks) should decide what to display after
        # successful delete
        return False

    def display(self) -> "DisplayResponse":
        self.emit("set_page_title", title="Delete {}".format(self.task.title))
        return self.render_template_string(
            "<h1>Delete {{task.title}}</h1>"
            """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
            """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
        )


@config(Component.Config(changes_url=False))
class PageTitle(Component):
    def __init__(self, title: str = ""):
        self._level: Optional[int] = None
        super().__init__()

    @listener(event="set_page_title")
    def on_set_page_title(self, event: Event):
        if self._level is None or (
            event.source and event.source._config.hiearchy_level > self._level
        ):
            self._level = event.source._config.hiearchy_level if event.source else 0
            self.state.title = event.title

    @action(deferred=True)
    def display(self):
        return self.render_template_string("<title>{{title}}</title>")


def inject_parent_and_wip_id(self: Component, component_config: ComponentConfig):
    return dict(parent_task_id=self.state.task_id, wip_id=self.state.wip_id)


@config(
    Component.Config(
        components=dict(
            page_title=PageTitle,
            tasks=(
                TaskList,
                TaskList.Config(
                    components=dict(
                        view=(
                            ViewTask,
                            ViewTask.Config(
                                components=dict(
                                    subtasks=(
                                        TaskList,
                                        TaskList.Config(components=dict(view=ViewTask)),
                                    )
                                ),
                                inject_into_components=inject_parent_and_wip_id,
                            ),
                        ),
                        edit=(
                            EditTask,
                            EditTask.Config(
                                components=dict(
                                    subtasks=(
                                        TaskList,
                                        TaskList.Config(
                                            components=dict(
                                                add=AddTask,
                                                edit=EditTask,
                                                delete=DeleteTask,
                                            )
                                        ),
                                    )
                                ),
                                inject_into_components=inject_parent_and_wip_id,
                            ),
                        ),
                        add=(
                            AddTask,
                            AddTask.Config(
                                components=dict(
                                    subtasks=(
                                        TaskList,
                                        TaskList.Config(
                                            components=dict(
                                                add=AddTask,
                                                edit=EditTask,
                                                delete=DeleteTask,
                                            ),
                                        ),
                                    )
                                ),
                                inject_into_components=inject_parent_and_wip_id,
                            ),
                        ),
                        delete=DeleteTask,
                    ),
                ),
            ),
        )
    ),
)
class TasksPage(Component):
    def __init__(self) -> None:
        super().__init__()
        self.unauthorized = False

    @listener(event="_exception")
    def on_exception(self, event):
        if isinstance(event.exception, Unauthorized):
            self.unauthorized = True
            event.handled = True
            return True  # Force redisplay

    def display(self) -> "DisplayResponse":
        return self.render_template_string(
            "<html><head>{{component('page_title')}}</head>"
            "<body>"
            "{% if unauthorized %}User is not authorized"
            "{% else %}{{component('tasks')}}"
            "{% endif %}"
            "</body></html>"
        )


def test_unauthorisied(jmb, client):
    global session, tasks_db
    jmb.add_page("tasks", TasksPage)

    # display unauthorized
    r = client.get("/tasks")
    assert r.status_code == 200
    assert r.data == (
        """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">"""
        "\n"
        """<html jmb-name="/tasks" jmb-data=\'{"actions":{},"changesUrl":true,"state":{},"url":"/tasks"}\'>"""
        """<head><title jmb-name="/tasks/page_title" jmb-data=\'{"actions":{},"changesUrl":false,"state":{"title":""},"url":"/tasks/page_title"}\'></title></head>"""
        """<body>"""
        "User is not authorized"
        """</body></html>"""
    ).encode("utf-8")


def test_empty_list(jmb, client):
    global session, tasks_db
    jmb.add_page("tasks", TasksPage)
    # login
    session["user"] = User("admin")

    # display empty list
    r = client.get("/tasks")
    assert r.status_code == 200
    assert r.data == (
        """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">"""
        "\n"
        """<html jmb-name="/tasks" jmb-data=\'{"actions":{},"changesUrl":true,"state":{},"url":"/tasks"}\'>"""
        """<head><title jmb-name="/tasks/page_title" jmb-data=\'{"actions":{},"changesUrl":false,"state":{"title":"Tasks"},"url":"/tasks/page_title"}\'>Tasks</title></head>"""
        """<body>"""
        """<div jmb-name="/tasks/tasks" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"mode":"list","parent_task_id":null,"wip_id":null},"url":"/tasks/tasks"}\'>"""
        """<button type="button" jmb-on:click="$jmb.component('add').display()">Add</button>"""
        "<table>"
        "<tr><th>Task</th><th>Actions</th></tr>"
        "</table>"
        """</div>"""
        """</body></html>"""
    ).encode("utf-8")


def test_add_task_x(jmb, client):
    global session, tasks_db
    jmb.add_page("tasks", TasksPage)
    # login
    session["user"] = User("admin")
    # display add page
    r = client.post(
        "/tasks/tasks",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Tasks"),),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="list", parent_task_id=None, wip_id=None),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/add",
                        initParams=dict(),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/add",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 4
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="Add task")
    assert json_response[0]["dom"] == ("""<title>Add task</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="add", parent_task_id=None, wip_id=None
    )
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/add"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/add"
    assert json_response[2]["state"] == dict(
        form=dict(title="", description=None, error=None),
        parent_task_id=None,
        task_id=-1,
        wip_id=1,
    )
    assert json_response[2]["dom"] == (
        "<div>"
        "<h1>New task</h1>"
        """<label>Title:"""
        """<input type="text" value="" """
        """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
        """<label>Description:"""
        """<input type="text" value="" """
        """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
        "<h2>Subtasks</h2>"
        """<div><template jmb-placeholder="/tasks/tasks/add/subtasks"></template></div>"""
        "</div>"
    )
    assert json_response[3]["execName"] == "/tasks/tasks/add/subtasks"
    assert json_response[3]["state"] == dict(mode="list")
    assert json_response[3]["dom"] == (
        "<div>"
        """<button type="button" jmb-on:click="$jmb.component('add').display()">Add</button>"""
        "<table>"
        "<tr><th>Task</th><th>Actions</th></tr>"
        "</table>"
        "</div>"
    )
    # add new task
    r = client.post(
        "/tasks/tasks/add",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Add task")),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="add", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/add",
                        state=dict(
                            form=dict(title="", description=None, error=None),
                            parent_task_id=None,
                            wip_id=1,
                            task_id=-1,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/add/subtasks", state=dict(mode="list"),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/add",
                        initParams=dict(
                            form=dict(
                                title="Task 1", description="First task", error=None
                            ),
                            parent_task_id=None,
                            wip_id=1,
                            task_id=-1,
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 4
    assert len(session["wipdbs"]) == 0
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="View Task 1")
    assert json_response[0]["dom"] == ("""<title>View Task 1</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="view", wip_id=None, parent_task_id=None
    )
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/view"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/view"
    assert json_response[2]["state"] == dict(task_id=1, wip_id=None)
    assert json_response[2]["dom"] == (
        """<h1><a href="#" jmb-on:click="$jmb.component('..').display()">Back</a> Task 1</h1>"""
        """<div>First task</div>"""
        "<h2>Sub tasks</h2>"
        """<div><template jmb-placeholder="/tasks/tasks/view/subtasks"></template></div>"""
    )
    assert json_response[3]["execName"] == "/tasks/tasks/view/subtasks"
    assert json_response[3]["state"] == dict(mode="list")
    assert json_response[3]["dom"] == (
        "<div><table><tr><th>Task</th><th>Actions</th></tr></table></div>"
    )
    assert session == dict(user=User(username="admin"), wipdbs={})
    assert tasks_db == {1: Task(1, "Task 1", "First task")}
    # TODO check state of every component


def test_add_second_task_x(client, jmb):
    """Calling save without properly created seted wipdb_id and task_id"""
    global session, tasks_db
    jmb.add_page("tasks", TasksPage)
    session = dict(user=User("admin"), wipdbs=dict())
    tasks_db = {1: Task(1, "Task 1", "First task")}
    r = client.post(
        "/tasks/tasks/add",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Add task")),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="add", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/add",
                        state=dict(
                            form=dict(title="", description=None, error=None),
                            parent_task_id=None,
                            wip_id=None,
                            task_id=None,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/add/subtasks", state=dict(mode="list"),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/add",
                        initParams=dict(
                            form=dict(
                                title="Task 2", description="Second task", error=None
                            ),
                            parent_task_id=None,
                            wip_id=None,
                            task_id=None,
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert session == dict(user=User(username="admin"), wipdbs={})
    assert tasks_db == {
        1: Task(1, "Task 1", "First task"),
        2: Task(2, "Task 2", "Second task"),
    }
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 4
    assert len(session["wipdbs"]) == 0
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="View Task 2")
    assert json_response[0]["dom"] == ("""<title>View Task 2</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="view", wip_id=None, parent_task_id=None
    )
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/view"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/view"
    assert json_response[2]["state"] == dict(task_id=2, wip_id=None)
    assert json_response[2]["dom"] == (
        """<h1><a href="#" jmb-on:click="$jmb.component('..').display()">Back</a> Task 2</h1>"""
        """<div>Second task</div>"""
        "<h2>Sub tasks</h2>"
        """<div><template jmb-placeholder="/tasks/tasks/view/subtasks"></template></div>"""
    )
    assert json_response[3]["execName"] == "/tasks/tasks/view/subtasks"
    assert json_response[3]["state"] == dict(mode="list")
    assert json_response[3]["dom"] == (
        "<div><table><tr><th>Task</th><th>Actions</th></tr></table></div>"
    )


def test_refresh_task_list_x(jmb, client):
    jmb.add_page("tasks", TasksPage)

    global session, tasks_db
    session = dict(user=User(username="admin"), wipdbs={})
    tasks_db = {
        1: Task(1, "Task 1", "First task"),
        2: Task(2, "Task 2", "Second task"),
    }
    # refresh tasks list
    r = client.post(
        "/tasks/tasks",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Tasks"),),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="list", parent_task_id=None, wip_id=None),
                    ),
                ],
                commands=[
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 1
    assert json_response[0]["execName"] == "/tasks/tasks"
    assert json_response[0]["state"] == dict(
        mode="list", parent_task_id=None, wip_id=None
    )
    assert json_response[0]["dom"] == (
        "<div>"
        """<button type="button" jmb-on:click="$jmb.component(\'add\').display()">Add</button>"""
        "<table>"
        "<tr><th>Task</th><th>Actions</th></tr>"
        "<tr>"
        "<td>"
        """<a href="/tasks/tasks/view/1" jmb-on:click="$jmb.component(\'view\',{\'task_id\':1}).display()">Task 1</a>"""
        "</td>"
        "<td>"
        """<a href="/tasks/tasks/edit/1" jmb-on:click="$jmb.component(\'edit\',{\'task_id\':1}).display()">edit</a>"""
        """<a href="/tasks/tasks/delete/1" jmb-on:click="$jmb.component(\'delete\',{\'task_id\':1}).display()">delete</a>"""
        "</td>"
        "</tr>"
        "<tr>"
        "<td>"
        """<a href="/tasks/tasks/view/2" jmb-on:click="$jmb.component(\'view\',{\'task_id\':2}).display()">Task 2</a>"""
        "</td>"
        "<td>"
        """<a href="/tasks/tasks/edit/2" jmb-on:click="$jmb.component(\'edit\',{\'task_id\':2}).display()">edit</a>"""
        """<a href="/tasks/tasks/delete/2" jmb-on:click="$jmb.component(\'delete\',{\'task_id\':2}).display()">delete</a>"""
        "</td>"
        "</tr>"
        "</table>"
        "</div>"
    )


def test_edit_first_task_x(jmb, client):
    jmb.add_page("tasks", TasksPage)

    global session, tasks_db
    session = dict(user=User(username="admin"), wipdbs={})
    tasks_db = {
        1: Task(1, "Task 1", "First task"),
        2: Task(2, "Task 2", "Second task"),
    }
    # display edit first task
    r = client.post(
        "/tasks/tasks",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Tasks"),),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="list", parent_task_id=None, wip_id=None),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/edit",
                        initParams=dict(task_id=1,),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/edit",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 4
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="Edit Task 1")
    assert json_response[0]["dom"] == ("""<title>Edit Task 1</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="edit", parent_task_id=None, wip_id=None
    )
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/edit"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/edit"
    assert json_response[2]["state"] == dict(
        form=dict(title="Task 1", description="First task", error=None),
        task_id=1,
        wip_id=1,
    )
    assert json_response[2]["dom"] == (
        "<div>"
        "<h1>Edit Task 1</h1>"
        """<label>Title:"""
        """<input type="text" value="Task 1" """
        """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
        """<label>Description:"""
        """<input type="text" value="First task" """
        """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
        "<h2>Subtasks</h2>"
        """<div><template jmb-placeholder="/tasks/tasks/edit/subtasks"></template></div>"""
        "</div>"
    )
    assert json_response[3]["execName"] == "/tasks/tasks/edit/subtasks"
    assert json_response[3]["state"] == dict(mode="list")
    assert json_response[3]["dom"] == (
        "<div>"
        """<button type="button" jmb-on:click="$jmb.component('add').display()">Add</button>"""
        "<table>"
        "<tr><th>Task</th><th>Actions</th></tr>"
        "</table>"
        "</div>"
    )

    assert len(session["wipdbs"]) == 1
    assert tasks_db == {
        1: Task(1, "Task 1", "First task"),
        2: Task(2, "Task 2", "Second task"),
    }
    # do edit second task
    r = client.post(
        "/tasks/tasks",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(
                        execName="/tasks/page_title", state=dict(title="Edit Task 1"),
                    ),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="edit", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/edit",
                        state=dict(
                            form=dict(
                                title="Task 1", description="First task", error=None
                            ),
                            task_id=1,
                            wip_id=1,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks", state=dict(mode="list"),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/edit",
                        initParams=dict(
                            form=dict(
                                title="Extended task 1",
                                description="Extended first task",
                                error=None,
                            ),
                            task_id=1,
                            wip_id=1,
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/edit",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 4
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="View Extended task 1")
    assert json_response[0]["dom"] == ("""<title>View Extended task 1</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="view", parent_task_id=None, wip_id=None
    )
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/view"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/view"
    assert json_response[2]["state"] == dict(task_id=1, wip_id=None)
    assert json_response[2]["dom"] == (
        """<h1><a href="#" jmb-on:click="$jmb.component('..').display()">Back</a> Extended task 1</h1>"""
        """<div>Extended first task</div>"""
        "<h2>Sub tasks</h2>"
        """<div><template jmb-placeholder="/tasks/tasks/view/subtasks"></template></div>"""
    )
    assert json_response[3]["execName"] == "/tasks/tasks/view/subtasks"
    assert json_response[3]["state"] == dict(mode="list")
    assert json_response[3]["dom"] == (
        "<div><table><tr><th>Task</th><th>Actions</th></tr></table></div>"
    )

    assert session == dict(user=User(username="admin"), wipdbs={})
    assert tasks_db == {
        1: Task(1, "Extended task 1", "Extended first task"),
        2: Task(2, "Task 2", "Second task"),
    }


def test_add_subtask_to_second_task_x(jmb, client):
    jmb.add_page("tasks", TasksPage)

    global session, tasks_db
    session = dict(user=User(username="admin"), wipdbs={1: WipDb()})
    tasks_db = {
        1: Task(1, "Extended task 1", "Extended first task"),
        2: Task(2, "Task 2", "Second task"),
    }
    # display add
    r = client.post(
        "/tasks/tasks/edit/1/subtasks",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(
                        execName="/tasks/page_title", state=dict(title="Edit Task 2"),
                    ),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="edit", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/edit",
                        state=dict(
                            form=dict(
                                title="Task 2", description="Second task", error=None
                            ),
                            task_id=2,
                            wip_id=1,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks", state=dict(mode="list"),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/edit/subtasks/add",
                        initParams=dict(),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/edit/subtasks/add",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="Add task")
    assert json_response[0]["dom"] == ("""<title>Add task</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks/edit/subtasks"
    assert json_response[1]["state"] == dict(mode="add")
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/edit/subtasks/add"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/edit/subtasks/add"
    assert json_response[2]["state"] == dict(
        form=dict(title="", description=None, error=None), task_id=-1
    )
    assert json_response[2]["dom"] == (
        "<div>"
        "<h1>New task</h1>"
        """<label>Title:"""
        """<input type="text" value="" """
        """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
        """<label>Description:"""
        """<input type="text" value="" """
        """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
        "</div>"
    )
    # add subtask
    r = client.post(
        "/tasks/tasks/edit/2/subtasks/add",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(execName="/tasks/page_title", state=dict(title="Add task"),),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="edit", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/edit",
                        state=dict(
                            form=dict(
                                title="Task 2", description="Second task", error=None
                            ),
                            task_id=2,
                            wip_id=1,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks", state=dict(mode="add"),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks/add",
                        state=dict(
                            form=dict(title="", description=None, error=None),
                            task_id=-1,
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/tasks/tasks/edit/subtasks/add",
                        initParams=dict(
                            form=dict(
                                title="Subtask 2.1",
                                description="First subtask of second task",
                                error=None,
                            ),
                            task_id=-1,
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/edit/subtasks/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="Edit Subtask 2.1")
    assert json_response[0]["dom"] == ("""<title>Edit Subtask 2.1</title>""")
    assert json_response[1]["execName"] == "/tasks/tasks/edit/subtasks"
    assert json_response[1]["state"] == dict(mode="edit")
    assert json_response[1]["dom"] == (
        """<template jmb-placeholder="/tasks/tasks/edit/subtasks/edit"></template>"""
    )
    assert json_response[2]["execName"] == "/tasks/tasks/edit/subtasks/edit"
    assert json_response[2]["state"] == dict(
        form=dict(
            title="Subtask 2.1", description="First subtask of second task", error=None
        ),
        task_id=-1,
    )
    assert json_response[2]["dom"] == (
        "<div>"
        "<h1>Edit Subtask 2.1</h1>"
        """<label>Title:"""
        """<input type="text" value="Subtask 2.1" """
        """jmb-on.change.deferred="$jmb.set('form.title', this.value)"></label>"""
        """<label>Description:"""
        """<input type="text" value="First subtask of second task" """
        """jmb-on.change.deferred="$jmb.set('form.description', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel')">Cancel</button>"""
        "</div>"
    )
    # Save second task so that newlly created subtask is saved
    r = client.post(
        "/tasks/tasks/edit/2",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/tasks", state=dict()),
                    dict(
                        execName="/tasks/page_title",
                        state=dict(title="Edit Subtask 2.1"),
                    ),
                    dict(
                        execName="/tasks/tasks",
                        state=dict(mode="edit", parent_task_id=None, wip_id=None),
                    ),
                    dict(
                        execName="/tasks/tasks/edit",
                        state=dict(
                            form=dict(
                                title="Task 2", description="Second task", error=None
                            ),
                            task_id=2,
                            wip_id=1,
                        ),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks", state=dict(mode="edit"),
                    ),
                    dict(
                        execName="/tasks/tasks/edit/subtasks/edit",
                        state=dict(
                            form=dict(
                                title="Subtask 2.1",
                                description="First subtask of second task",
                                error=None,
                            ),
                            task_id=-1,
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="call",
                        componentExecName="/tasks/tasks/edit",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    assert len(session["wipdbs"]) == 0
    assert tasks_db == {
        1: Task(1, "Extended task 1", "Extended first task"),
        2: Task(2, "Task 2", "Second task"),
        3: Task(3, "Subtask 2.1", "First subtask of second task", parent_id=2),
    }
    json_response = json.loads(r.data)
    assert json_response[0]["execName"] == "/tasks/page_title"
    assert json_response[0]["state"] == dict(title="View Task 2")
    assert json_response[1]["execName"] == "/tasks/tasks"
    assert json_response[1]["state"] == dict(
        mode="view", parent_task_id=None, wip_id=None
    )
    assert json_response[2]["execName"] == "/tasks/tasks/view"
    assert json_response[2]["state"] == dict(task_id=2, wip_id=None)
    assert json_response[3]["execName"] == "/tasks/tasks/view/subtasks"
    assert json_response[3]["state"] == dict(mode="list")


# TODO edit second task and add subtasks (x-jembe)
# TODO add subtask with two level subtasks (x-jembe)
# TODO edit task and add, edit and delete subtasks in bulk (x-jembe)
# TODO delete task



from typing import TYPE_CHECKING, Union, Tuple, Type, Dict, Any, get_args, get_origin
import re
import collections
import inspect
from abc import ABC, abstractmethod
from importlib import import_module
from dataclasses import is_dataclass, asdict
from datetime import date, datetime
from flask import Response, json, current_app
from jembe.exceptions import JembeError

if TYPE_CHECKING:  # pragma: no cover
    import jembe

ComponentRef = Union[
    Union[Type["jembe.Component"], str],
    Tuple[
        Union[Type["jembe.Component"], str],
        Union["jembe.ComponentConfig", Dict[str, Any]],
    ],
]

DisplayResponse = Union[str, Response]


def exec_name_to_full_name(exec_name: str) -> str:
    """
    Removes component keys from exec name to get full_name.

    keys in exec_name are separated by . (dot)
    """
    return "/".join(ck.split(".")[0] for ck in exec_name.split("/"))


def is_page_exec_name(exec_name: str) -> bool:
    """Is Exec name name of the page"""
    return len(exec_name.strip("/").split("/")) == 1


def direct_child_name(
    component: "jembe.Component", subcompoenent_full_name: str
) -> str:
    """
    Returns name of the direct child of component from subcomponent.
    If subcompoennt is not under component raise error.
    """
    parent_names = component._config.full_name.split("/")
    child_names = subcompoenent_full_name.split("/")
    if parent_names != child_names[: len(parent_names)]:
        raise ValueError(
            f"Component {subcompoenent_full_name} is not subcomponent of {component._config.full_name}"
        )
    return child_names[len(parent_names)]


def is_direct_child_name(exec_name: str, sub_exec_name: str) -> bool:
    """Returns True if sub_exec_name is direct child of exec_name"""
    if not sub_exec_name.startswith(exec_name) or exec_name == sub_exec_name:
        return False
    relative_name = sub_exec_name[len(exec_name) :]
    if relative_name[0] != "/":
        return False
    if len(relative_name.split("/")) > 2:
        return False
    return True


def is_child_name(exec_name: str, sub_exec_name: str) -> bool:
    """Returns True if sub_exec_name is under exec_name"""
    if not sub_exec_name.startswith(exec_name) or exec_name == sub_exec_name:
        return False
    relative_name = sub_exec_name[len(exec_name) :]
    if relative_name[0] != "/":
        return False
    return True


def parent_exec_name(exec_name: str) -> str:
    return "/".join(exec_name.split("/")[:-1])


def import_by_name(object_name: str) -> Any:
    try:
        str_split = object_name.split(".")
        object_name = str_split[-1]
        modul_name = ".".join(str_split[:-1])
        modul = import_module(modul_name)
        object_type = getattr(modul, object_name)
    except:
        raise ValueError(f"Import with name {object_name} doesn't exist")
    return object_type


class JembeInitParamSupport(ABC):
    @classmethod
    @abstractmethod
    def dump_init_param(cls, value: Any) -> Any:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def load_init_param(cls, value: Any) -> Any:
        raise NotImplementedError()


# Transforming python type to json serializable type back and forth
def dump_param(annotation, value: Any) -> Any:
    """
    Transforms annotated component state param into json serializable type ready to be
    used by jembe client javascript, either as state variable, url query variale,
    jrl etc.
    """
    source_type, can_be_none = get_annotation_type(annotation)
    if value is None and can_be_none:
        return None
    elif _eq_type(source_type, list, tuple, set, collections.abc.Sequence):
        targs = get_args(source_type)
        if len(targs) == 0:
            _dump_load_unspecified_warning(source_type)
            return value
        return list(dump_param(targs[0], item) for item in value)
    elif _eq_type(source_type, dict):
        targs = get_args(source_type)
        if len(targs) < 2:
            _dump_load_unspecified_warning(source_type)
            return value
        return {
            dump_param(targs[0], k): dump_param(targs[1], v) for k, v in value.items()
        }
    elif (
        source_type == JembeInitParamSupport
        or (
            inspect.isclass(source_type)
            and issubclass(source_type, JembeInitParamSupport)
        )
        or (
            inspect.isclass(get_origin(source_type))
            and issubclass(
                get_origin(source_type), JembeInitParamSupport  # type:ignore
            )
        )
        or isinstance(value, JembeInitParamSupport)
    ):
        return source_type.dump_init_param(value)
    elif _eq_type(source_type, int, float, bool, str):
        return value
    elif _eq_type(source_type, date, datetime):
        return value.isoformat()
    elif is_dataclass(source_type):
        return asdict(value)
    elif source_type == Any:
        _dump_load_unspecified_warning(source_type)
        return value
    raise JembeError(f"Unsupported state/init param type {source_type}:{value}")


def load_param(annotation, value: Any) -> Any:
    """
    Loads component sate params received by jembe client javascript
    and transform it to python type according to its annotation
    ready to be process by component
    """
    target_type, can_be_none = get_annotation_type(annotation)
    if value is None and can_be_none:
        return None
    if _eq_type(target_type, list):
        if isinstance(value, str):
            value = _decode_str_to_type(list, value)
        targs = get_args(target_type)
        if len(targs) == 0:
            _dump_load_unspecified_warning(target_type)
            return value
        return list(load_param(targs[0], v) for v in value)
    elif _eq_type(target_type, tuple, collections.abc.Sequence):
        if isinstance(value, str):
            value = _decode_str_to_type(list, value)
        targs = get_args(target_type)
        if len(targs) == 0:
            _dump_load_unspecified_warning(target_type)
            return value
        return tuple([load_param(targs[0], v) for v in value])
    elif _eq_type(target_type, set):
        if isinstance(value, str):
            value = _decode_str_to_type(list, value)
        targs = get_args(target_type)
        if len(targs) == 0:
            _dump_load_unspecified_warning(target_type)
            return value
        return set([load_param(targs[0], v) for v in value])
    elif _eq_type(target_type, dict):
        if isinstance(value, str):
            value = _decode_str_to_type(dict, value)
        targs = get_args(target_type)
        if len(targs) == 0:
            _dump_load_unspecified_warning(target_type)
            return value
        return {
            load_param(targs[0], k): load_param(targs[1], v) for k, v in value.items()
        }
    elif (
        target_type == JembeInitParamSupport
        or (
            inspect.isclass(target_type)
            and issubclass(target_type, JembeInitParamSupport)
        )
        or (
            inspect.isclass(get_origin(target_type))
            and issubclass(
                get_origin(target_type), JembeInitParamSupport  # type:ignore
            )
        )
    ):
        return target_type.load_init_param(value)
    elif _eq_type(target_type, int):
        return int(value)
    elif _eq_type(target_type, float):
        return float(value)
    elif _eq_type(target_type, bool):
        if isinstance(value, str):
            value = _decode_str_to_type(dict, value)
        return bool(value)
    elif _eq_type(target_type, str):
        return str(value)
    elif _eq_type(target_type, date):
        return date.fromisoformat(value)
    elif _eq_type(target_type, datetime):
        return datetime.fromisoformat(value)
    elif is_dataclass(target_type):
        return dataclass_from_dict(target_type, value)
    elif target_type == Any:
        _dump_load_unspecified_warning(Any)
        return value
    raise JembeError(f"Unsupported state/init param type {target_type}:{value}")


def _eq_type(checked_type: Type[Any], *compare_with: Type[Any]) -> bool:
    for cw in compare_with:
        if checked_type == cw or get_origin(checked_type) == cw:
            return True
    return False


def get_annotation_type(annotation):
    """returns tuple(annotation_type, is_optional)"""

    def _geta(annotation):
        if inspect.isfunction(annotation):
            # handling typing.NewType
            # TODO make it robust and to cover all cases
            return annotation.__supertype__
        else:
            return annotation

    if get_origin(annotation) is Union and type(None) in get_args(annotation):
        # is_optional
        return (_geta(get_args(annotation)[0]), True)
    else:
        return (_geta(annotation), False)


def dataclass_from_dict(klass, dikt):
    try:
        fieldtypes = klass.__annotations__
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except AttributeError:
        if isinstance(dikt, (tuple, list)):
            return [dataclass_from_dict(klass.__args__[0], f) for f in dikt]
        return dikt


def _decode_str_to_type(ttype, value: str):
    if ttype == bool:
        return ttype(value.upper() == "TRUE")

    return ttype(json.loads(re.sub("(?<!\\\\)'", '"', value)))


def _dump_load_unspecified_warning(ttype):
    current_app.logger.warning(
        "Do not use '{ttype}' annotation for component state params "
        "you should be more specific in other to enable proper transformation"
        "into and out json"
    )



from .app import (
    Jembe,
    Processor,
    get_jembe,
    get_storage,
    get_storages,
    get_private_storage,
    get_public_storage,
    get_temp_storage,
)
from .component import Component, ComponentReference
from .common import JembeInitParamSupport, ComponentRef, DisplayResponse
from .component_config import config, action, listener, redisplay, UrlPath, ComponentConfig, RedisplayFlag
from .processor import Event
from .exceptions import (
    ComponentPreviousStateUnavaiableError,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    AccessDenied,
    Conflict,
    Gone,
    NotImplemented,
    InternalServerError,
    JembeError
)
from .files import File, Storage, DiskStorage
from .utils import page_url, run_only_once

__all__ = (
    "Jembe",
    "Processor",
    "get_jembe",
    "get_storage",
    "get_storages",
    "get_private_storage",
    "get_public_storage",
    "get_temp_storage",
    "Component",
    "ComponentReference",
    "JembeInitParamSupport",
    "action",
    "listener",
    "redisplay",
    "config",
    "UrlPath",
    "ComponentConfig",
    "RedisplayFlag",
    "ComponentRef",
    "DisplayResponse",
    "Event",
    "ComponentPreviousStateUnavaiableError",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "AccessDenied",
    "Conflict",
    "Gone",
    "NotImplemented",
    "InternalServerError",
    "JembeError",
    "File",
    "Storage",
    "DiskStorage",
    "page_url",
    "run_only_once",
)
